/* *****************************************************************
 *
 *  $Id: f0m_jml_2_x.c,v 1.1 2002/09/27 12:43:01 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <03/03/26 23:11:57 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "glob.h"
#include "chain.h"
#include "msg.h"
#include "f0s_tkout.h"
#include "f0m_tkin.h"
#include "f0m_cmd_codes.h"
#include "f0m_jml_2_x.h"


/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */
#define MAX_BUF 64*1024

#define STRLEN(s)		(s) == NULL ? 0 : strlen (s)
#define STRING(s)		(s) == NULL ? "" : (s)


/* Used in "f0m_jml_2_tkin()" */
#define SKIP(c,n)		p = strchr (p, (c)); \
						assert (p != NULL);  \
						p += (n)
#define NL				SKIP ('\n', 0); while (*++p == ' ')
#define EQ				SKIP ('=', 2)
#define SP				SKIP (' ', 1)

#define TEST(s)			(strncmp (p, (s), STRLEN (s)) == 0)
#define CMP(s)			{int e = strncmp (p, (s), STRLEN (s)); assert (e == 0);}
#define INT				(int)strtol (p, &p, 0)

#define X_I(n)			i[n] = INT
#define X_S(n)									\
		do {									\
			s[(n)].l = INT;						\
			SP;									\
			if (s[(n)].l == 0)					\
				s[(n)].c = NULL;				\
			else {								\
				char c = *(p + (s[(n)].l));		\
				*(p + (s[(n)].l)) = '\0';		\
				s[n].c = strdup (p);			\
				*(p + (s[(n)].l)) = c;			\
				p += s[(n)].l;					\
			}									\
		} while (0)




#define X_II(i)			i = INT
#define X_AA(a)			a = (void *)INT
#define X_SS(s)									\
		do {									\
			int len = INT;						\
		    if (len == 0)						\
				(s) = NULL;						\
			else {								\
				char c;			\
				SP;									\
				c = *(p + len);			\
				*(p + len) = '\0';				\
				(s) = strdup (p);				\
				*(p + len) = c;					\
				p += len;						\
			}									\
		} while (0)


typedef struct ch_data_s ch_data_t;
struct ch_data_s {
	char  **str;
	int     code;
	data_t *data;
};



/* #################################################################
 *		PRIVATE functions
 * ################################################################# */



PRIVATE chain_t *extract_chain (void *(*f)(), const char *var,
								ch_data_t *zz)
{
	int n_elems;
	void *elem;
	chain_t *ch;
	char *p;

	p = *zz->str;
	ch = chain_create ();
	assert (ch != NULL);
	CMP ("chain"); SKIP ('[', 1);
	X_II (n_elems);
	SKIP (':', 1);
	CMP (var);
	NL;
	while (n_elems--) {
		*zz->str = p;
		elem = f (zz);
		p = *zz->str;
		assert (elem != NULL);
		chain_enqueue (ch, elem);
	}
	NL;
	*zz->str = p;
	return ch;
}



PRIVATE void *extract_node (ch_data_t *zz)
{
	void *base;
	char *p = *zz->str;

#define NEW(type) type *d; base = (type *)malloc (sizeof (*d)); d = base

 	switch (zz->code) {
	case R_TIDS:
	{   NEW(rtids_t);
		CMP ("tid"); NL;
		CMP ("gtid"); EQ; X_II (d->gtid); NL;
		CMP ("tp_pid"); EQ; X_II (d->tp_pid); NL;
		CMP ("dbg_pid"); EQ; X_II (d->dbg_pid); NL;
		CMP ("attached"); EQ; X_II (d->attached); NL;
		CMP ("ltid"); EQ; X_II (d->ltid); NL;
		CMP ("machine"); EQ; X_SS(d->machine); NL;
		NL;
		break;
	}
	case R_CLIENTS:
	{   NEW(rclients_t);
		CMP ("clients"); NL;
		CMP ("pid"); EQ; X_II (d->pid); NL;
		CMP ("filename"); EQ; X_SS(d->filename); NL;
		CMP ("hostaddr"); EQ; X_SS(d->hostaddr); NL;
		NL;
		break;
	}
	case R_KILL:
	{   /* NEW(rkill_t); */
		CMP ("kill"); NL;
		/* Empty */
		NL;
		break;
	}
	case R_ADD_DEBUGGER:
	{   NEW(radddebugger_t);
		CMP ("add_debugger"); NL;
		CMP ("tid"); EQ; X_II(d->tid); NL;
		NL;
		break;
	}
	case R_ATTACH:
	{   NEW(rattach_t);
		CMP ("attach"); NL;
		CMP ("tid"); EQ; X_II(d->tid); NL;
		CMP ("tp_pid"); EQ; X_II(d->tp_pid); NL;
		NL;
		break;
	}
	case R_DETACH:
	{   NEW(rdetach_t);
		CMP ("detach"); NL;
		CMP ("tp_pid"); EQ; X_II(d->tp_pid); NL;
		NL;
		break;
	}
	case R_SYMBOL_FILE:
	{   /* NEW(rsymbolfile_t); */
		CMP ("symbol_file"); NL;
		/* Empty */
/* 		CMP ("file"); EQ; X_SS(d->file); NL; */
		NL;
		break;
	}
	case R_FILE:
	{   NEW(rfile_t);
		CMP ("file"); NL;
		CMP ("tid"); EQ; X_II(d->tid); NL;
/* 		CMP ("file"); EQ; X_SS(d->file); NL; */
		NL;
		break;
	}
	case R_RUN:
	{   NEW(rrun_t);
		CMP ("run"); NL;
		CMP ("file"); EQ; X_SS(d->file); NL;
		NL;
		break;
	}
	case R_STEP:
	{   /* NEW(rstep_t); */
		CMP ("step"); NL;
		/* Empty */
		NL;
		break;
	}
	case R_NEXT:
	{   /* NEW(rnext_t); */
		CMP ("next"); NL;
		/* Empty */
		NL;
		break;
	}
	case R_CONTINUE:
	{   /* NEW(rcontinue_t); */
		CMP ("continue"); NL;
		/* Empty */
		NL;
		break;
	}
	case R_FINISH:
	{   NEW(rfinish_t);
		CMP ("finish"); NL;
		CMP ("s_address"); EQ; X_AA(d->s_address); NL;
		CMP ("s_function"); EQ; X_SS(d->s_function); NL;
		CMP ("s_args"); EQ; X_SS(d->s_args); NL;
		CMP ("s_file"); EQ; X_SS(d->s_file); NL;
		CMP ("s_line"); EQ; X_II(d->s_line); NL;
		CMP ("dbg_var"); EQ; X_II(d->dbg_var); NL;
		CMP ("value"); EQ; X_SS(d->value); NL;
		NL;
		break;
	}
	case R_CALL_VOID:
	{   /* NEW(rcallvoid_t); */
		CMP ("call_void"); NL;
		/* Empty */
		NL;
   		break;
	}
	case R_BREAK:
	{   NEW(rbreak_t);
		CMP ("break"); NL;
		CMP ("bpid"); EQ; X_II(d->bpid); NL;
		CMP ("address"); EQ; X_AA(d->address); NL;
		CMP ("file"); EQ; X_SS(d->file); NL;
		CMP ("line"); EQ; X_II(d->line); NL;
		NL;
		break;
	}
	case R_DELETE:
	{   /* NEW(rdelete_t); */
		CMP ("delete"); NL;
		/* Empty */
		NL;
		break;
	}
	case R_SET_VARIABLE:
	{   /* NEW(rsetvariable_t); */
		CMP ("set_variable"); NL;
		/* Empty */
		NL;
		break;
	}
	case R_EVALUATE:
	{   NEW(revaluate_t);
		CMP ("evaluate"); NL;
		CMP ("dbg_var"); EQ; X_II(d->dbg_var); NL;
		CMP ("value"); EQ; X_SS(d->value); NL;
		NL;
		break;
	}
	case R_DISPLAY_EXPRESSION:
	{   /* NEW(rdisplayexpression_t); */
		CMP ("display_expression"); NL;
		/* Empty */
		NL;
		break;
	}
	case R_UNDISPLAY:
	{   /* NEW(rundisplay_t); */
		CMP ("undisplay"); NL;
		/* Empty */
		NL;
		break;
	}
	case R_LIST:
	{   NEW(rlist_t);
		CMP ("list"); NL;
		CMP ("line_number"); EQ; X_II (d->line_number); NL;
		CMP ("line_text"); EQ; X_SS (d->line_text); NL;
		NL;
		break;
	}
	case R_INFO_LOCALS:
	{   NEW(rinfolocals_t);
		CMP ("info_locals_args"); NL;
		CMP ("var"); EQ; X_SS(d->var); NL;
		CMP ("value"); EQ; X_SS(d->value); NL;
		NL;
		break;
	}
	case R_INFO_LINE_NO_CODE:
	case R_INFO_LINE:
	{   NEW(rinfoline_t);
		CMP ("info_line"); NL;
		CMP ("line"); EQ; X_II(d->line); NL; 
		CMP ("file"); EQ; X_SS(d->file); NL; 
		CMP ("start_function"); EQ;
		X_SS(d->start_function); NL;
		CMP ("start_function_offset"); EQ;
		X_II(d->start_function_offset); NL;
		CMP ("end_function"); EQ;
		X_SS(d->end_function); NL;
		CMP ("end_function_offset"); EQ;
		X_II(d->end_function_offset); NL;
		CMP ("start_address"); EQ;
		X_AA(d->start_address); NL;
		CMP ("end_address"); EQ;
		X_AA(d->end_address); NL;
		NL;
		break;
	}
	case R_INFO_PROGRAM:
	{   NEW(rinfoprogram_t);
		CMP ("info_program"); NL;
		CMP ("tp_pid"); EQ; X_II(d->tp_pid); NL;
		CMP ("address"); EQ;
		X_AA(d->address); NL;
		CMP ("why"); EQ; X_SS(d->why); NL;
		CMP ("bpid"); EQ; X_II(d->bpid); NL;
		CMP ("signal"); EQ; X_SS(d->signal); NL;
		NL;
		break;
	}
	case R_INFO_STACK:
	{   NEW(rframe_t);
		CMP ("info_stack"); NL;
		CMP ("frame"); EQ; X_II (d->frame); NL;
		CMP ("address"); EQ; X_AA(d->address); NL;
		CMP ("function"); EQ; X_SS (d->function); NL;
		CMP ("arguments"); EQ; X_SS (d->arguments); NL;
		CMP ("file"); EQ; X_SS (d->file); NL;
		CMP ("line"); EQ; X_II (d->line); NL;
		NL;
		break;
	}
	case R_INFO_BREAK:
	{   NEW(rinfobreak_t);
		CMP ("info_break"); NL;
		CMP ("bpid"); EQ; X_II (d->bpid); NL;
		CMP ("type"); EQ; X_SS (d->type); NL;
		CMP ("disp"); EQ; X_SS (d->disp); NL;
		CMP ("enabled"); EQ; X_II (d->enabled); NL;
		CMP ("address"); EQ; X_AA (d->address); NL;
		CMP ("function"); EQ; X_SS (d->function); NL;
		CMP ("file"); EQ; X_SS (d->file); NL;
		CMP ("line"); EQ; X_II (d->line); NL;
		CMP ("hit_counter"); EQ; X_II (d->hit_counter); NL;
		CMP ("ign_counter"); EQ; X_II (d->ign_counter); NL;
		CMP ("condition"); EQ; X_SS (d->condition); NL;
		NL;
		break;
	}
	case R_INFO_DISPLAY:
	{   NEW(rdisplay_t);
		CMP ("info_display"); NL;
		CMP ("dispid"); EQ; X_II (d->dispid); NL;
		CMP ("enabled"); EQ; X_II (d->enabled); NL;
		CMP ("expression"); EQ; X_SS (d->expression); NL;
		CMP ("value"); EQ; X_SS (d->value); NL;
		NL;
		break;
	}
	case R_INFO_THREADS:
	{   NEW(rinfothreads_t);
		CMP ("info_threads"); NL;
		CMP ("active"); EQ; X_II (d->active); NL;
		CMP ("th_id"); EQ; X_II (d->th_id); NL;
		CMP ("tp_pid"); EQ; X_II (d->tp_pid); NL;
		CMP ("frame"); EQ; X_AA (d->frame); NL;
		CMP ("function"); EQ; X_SS (d->function); NL;
		CMP ("arguments"); EQ; X_SS (d->arguments); NL;
		CMP ("file"); EQ; X_SS (d->file); NL;
		CMP ("line"); EQ; X_II (d->line); NL;
		NL;
		break;
	}
	case R_TTY:
	{   /* NEW(rtty_t); */
		CMP ("tty"); NL;
		/* Empty */
		NL;
		break;
	}
	case R_THREAD:
	{   /* NEW(rthread_t); */
		CMP ("thread"); NL;
		/* Empty */
		NL;
		break;
	}
	case R_SENDTO:
	{   NEW(rsendto_t);
		CMP ("sendto"); NL;
		d->output = zz->data->str;
		NL;
   		break;
	}
	case R_REGISTER_CLIENT:
	{   NEW(rclients_t);
		CMP ("register_client"); NL;
		CMP ("pid"); EQ; X_II (d->pid); NL;
		CMP ("filename"); EQ; X_SS(d->filename); NL;
		CMP ("hostaddr"); EQ; X_SS(d->hostaddr); NL;
		NL;
		break;
	}
	case O_DISPLAY:
	{   NEW(odisplay_t);
		CMP ("display"); NL;
		CMP ("dispid"); EQ; X_II (d->dispid); NL;
		CMP ("enabled"); EQ; X_II (d->enabled); NL;
		CMP ("expression"); EQ; X_SS (d->expression); NL;
		CMP ("value"); EQ; X_SS (d->value); NL;
		NL;
		break;
	}
	case O_BPHIT:
	{   NEW(obphit_t);
		CMP ("bp_hit"); NL;
		CMP ("bpid"); EQ; X_II(d->bpid); NL;
		CMP ("function"); EQ; X_SS (d->function); NL;
		CMP ("args"); EQ; X_SS (d->args); NL;
		CMP ("file"); EQ; X_SS (d->file); NL;
		CMP ("line"); EQ;  X_II (d->line); NL;
		NL;
		break;
	}
	case O_SIGNAL:
	{   NEW(osignal_t);
		CMP ("signal"); NL;
		CMP ("name"); EQ; X_SS (d->name); NL;
		NL;
		break;
	}
	case O_NEW_THREAD:
	{   NEW(onewthread_t);
		CMP ("new_thread"); NL;
		CMP ("id"); EQ; X_II(d->id); NL;
		NL;
		break;
	}
	case O_SWITCH_TO_THREAD:
	{   NEW(othread_t);
		CMP ("thread"); NL;
		CMP ("pid"); EQ;  X_II (d->pid); NL;
		NL;
		break;
	}
	case P_POSITION:
	{   NEW(pposition_t);
		CMP ("position"); NL;
		CMP ("file"); EQ; X_SS (d->file); NL;
		CMP ("line"); EQ;  X_II (d->line); NL;
		CMP ("pos"); EQ;  X_II (d->pos); NL;
		CMP ("where"); EQ; X_SS (d->where); NL;
		CMP ("frame"); EQ; X_AA (d->frame); NL;
		NL;
		break;
	}
	default:
		msg_pid_fatal ("\n\
#################################################################\n\
%s:%d:extract_node()---Unexpected code = %d\n\
#################################################################\n",
                        __FILE__, __LINE__, zz->code);
		break;
	}
	*zz->str = p;
	return base;
}


PRIVATE void *jml_2_oob (ch_data_t *zz)
{
	char *p;
	data_t *oob = NULL;

	p = *zz->str;
	CMP ("oob");				/* Check it's a "oobdata_t" */
	NL;							/* Skip line */
	if (! TEST ("}")) {
		oob = f0s_tkout_data_new ();

		/* The 'oobdata_t' was not empty */
		CMP ("str"); EQ; X_SS(oob->str);
		NL;
	
		CMP ("code"); EQ; X_II(oob->code);
		NL;

		zz->code = oob->code;
		zz->data = NULL;
		zz->str = &p;
		oob->data = extract_chain (extract_node, "data", zz);
	}
	NL;
	*zz->str = p;
	return oob;
}


PRIVATE data_t *jml_2_data (const char *varname, char **p1)
{
	ch_data_t zz;
	data_t *var = NULL;
	
	char *p;

	p = *p1;
	CMP (varname);				/* Check it's a "data_t" */
	NL;							/* Skip line */
	if (! TEST ("}")) {
		var = f0s_tkout_data_new ();

		/* The 'data_t' was not empty */
		CMP ("str"); EQ;
		X_SS(var->str);				/* s[0] = string */
		NL;
	
		CMP ("code"); EQ;			/* Check next field it's "code" */
		X_II(var->code);				/* i[0] = code */
		NL;

		zz.code = var->code;
		zz.str = &p;
		zz.data = var;
		var->data = extract_chain (extract_node, "data", &zz);
		p = *zz.str;
	}	
	NL;
	*p1 = p;
	return var;
}





/* #################################################################
 *		PUBLIC functions
 * ################################################################# */


/* -----------------------------------------------------------------
 * Function:    tkin_t *f0m_jml_2_tkin (const char *str)
 *
 * Description: Convert a string into a 'tkin_t' strucuture
 *
 * Returns:     TheA dynamically allocated 'tkin_t' strucuture
 */

PUBLIC tkin_t *f0m_jml_2_tkin (const char *str)
{
	tkin_t *tk;
	int code;
	int tid;
	struct {
		int   l;
		char *c;
	} s[3];						/* For strings */
	int i[3];					/* For integers */
	
	char *p;

	p = (char *)str;
	CMP ("tkin");				/* Check it's a "tkin" */
	NL;							/* Skip line */
	CMP ("tid");				/* Check next field it's "code" */
	EQ;							/* Skip till number */
	tid = INT;					/* Extract code form string */
	assert (tid == -1  ||  tid > 0);
	NL;
	CMP ("code");				/* Check next field it's "code" */
	EQ;							/* Skip till number */
	code = INT;					/* Extract code form string */
	assert (code > 0);
	NL;
 	switch (code) {
 	case C_TIDS:
		CMP ("tids"); NL;		/* Verify main field */
		CMP ("options"); EQ;	/* Verify sub-field name */
		X_S(0);
		NL;						/* Skip "\t}\n" */
		tk = f0m_tkin_new (tid, C_TIDS, s[0].c);
		break;
 	case C_CLIENTS:
		CMP ("clients"); NL;		/* Verify main field */
		CMP ("options"); EQ; X_S(0); NL;
		NL;
		tk = f0m_tkin_new (tid, C_CLIENTS, s[0].c);
		break;
	case C_KILL:
 		CMP ("kill"); NL;
		NL;
		tk = f0m_tkin_new (tid, C_KILL);
		break;
	case C_ADD_DEBUGGER:
 		CMP ("add_debugger"); NL;
		CMP ("host"); EQ; X_S(0); NL;
		CMP("dbg_pid"); EQ; X_I(0); NL;
		CMP("tp_pid"); EQ; X_I(1); NL;
		CMP ("cmd_channel"); EQ; X_S(1); NL;
		CMP ("rep_channel"); EQ; X_S(2); NL;
		NL;
		tk = f0m_tkin_new (tid, C_ADD_DEBUGGER,
						   s[0].c, i[0], i[1], s[1].c, s[2].c);
		break;
	case C_ATTACH:
 		CMP ("attach"); NL;
		CMP("tp_pid"); EQ; X_I(0); NL;
		CMP ("host"); EQ; X_S(0); NL;
		NL;
		tk = f0m_tkin_new (tid, C_ATTACH, i[0], s[0].c);
		break;
	case C_DETACH:
 		CMP ("detach"); NL;
		NL;
		tk = f0m_tkin_new (tid, C_DETACH);
		break;
	case C_SYMBOL_FILE:
 		CMP ("symbol_file"); NL;
		CMP ("file"); EQ; X_S(0); NL;
		NL;
		tk = f0m_tkin_new (tid, C_SYMBOL_FILE, s[0].c);
		break;
	case C_FILE:
 		CMP ("file"); NL;
		CMP ("file"); EQ; X_S(0); NL;
		CMP ("host"); EQ; X_S(1); NL;
		NL;
		tk = f0m_tkin_new (tid, C_FILE, s[0].c, s[1].c);
		break;
	case C_RUN:
 		CMP ("run"); NL;
		CMP ("args"); EQ; X_S(0); NL;
		NL;
		tk = f0m_tkin_new (tid, C_RUN, s[0].c);
		break;
	case C_STEP:
 		CMP ("step"); NL;
		NL;
		tk = f0m_tkin_new (tid, C_STEP);
		break;
	case C_NEXT:
 		CMP ("next"); NL;
		NL;
		tk = f0m_tkin_new (tid, C_NEXT);
		break;
	case C_CONTINUE:
 		CMP ("continue"); NL;
		NL;
		tk = f0m_tkin_new (tid, C_CONTINUE);
		break;
	case C_FINISH:
 		CMP ("finish"); NL;
		NL;
		tk = f0m_tkin_new (tid, C_FINISH);
		break;
	case C_SIGNAL:
		CMP ("signal"); NL;
		CMP ("tid"); EQ; X_I(0); NL;
		CMP ("signum"); EQ; X_I(1); NL;
		NL;
		tk = f0m_tkin_new (tid, C_SIGNAL, i[0], i[1]);
		break;
	case C_BREAK:
 		CMP ("break"); NL;
		CMP ("file"); EQ; X_S(0); NL;
		CMP ("line"); EQ; X_I(0); NL;
		CMP ("function"); EQ; X_S(1); NL;
		NL;
		tk = f0m_tkin_new (tid, C_BREAK, s[0].c, i[0], s[1].c);
		break;
	case C_DELETE:
 		CMP ("delete"); NL;
		CMP ("bpid"); EQ; X_I(0); NL;
		NL;
		tk = f0m_tkin_new (tid, C_DELETE, i[0]);
		break;
	case C_SET_VARIABLE:
 		CMP ("set_variable"); NL;
		CMP ("variable"); EQ; X_S(0); NL;
		CMP ("expression"); EQ; X_S(1); NL;
		NL;
		tk = f0m_tkin_new (tid, C_SET_VARIABLE, s[0].c, s[1].c);
		break;
	case C_EVALUATE:
 		CMP ("evaluate"); NL;
		CMP ("expression"); EQ; X_S(0); NL;
		NL;
		tk = f0m_tkin_new (tid, C_EVALUATE, s[0].c);
		break;
	case C_DISPLAY:
 		CMP ("display"); NL;
		CMP ("expression"); EQ; X_S(0); NL;
		NL;
		tk = f0m_tkin_new (tid, C_DISPLAY, s[0].c);
		break;
	case C_UNDISPLAY:
		CMP ("undisplay"); NL;
		CMP ("dispid"); EQ; X_I(0); NL;
		NL;
		tk = f0m_tkin_new (tid, C_UNDISPLAY, i[0]);
		break;
	case C_CALL:
 		CMP ("call"); NL;
		CMP ("expression"); EQ; X_S(0); NL;
		NL;
		tk = f0m_tkin_new (tid, C_CALL, s[0].c);
		break;
	case C_LIST:
 		CMP ("list"); NL;
		CMP ("file"); EQ; X_S(0); NL;
		CMP ("start_line"); EQ; X_I(0); NL;
		CMP ("function"); EQ; X_S(1); NL;
		CMP ("n_lines"); EQ; X_I(1); NL;
		NL;
		tk = f0m_tkin_new (tid, C_LIST, s[0].c, i[0], s[1].c, i[1]);
		break;
	case C_TTY:
 		CMP ("tty"); NL;
		CMP ("file"); EQ; X_S(0); NL;
		NL;
		tk = f0m_tkin_new (tid, C_TTY, s[0].c);
		break;
	case C_UP:
 		CMP ("up"); NL;
		NL;
		tk = f0m_tkin_new (tid, C_UP);
		break;
	case C_DOWN:
 		CMP ("down"); NL;
		NL;
		tk = f0m_tkin_new (tid, C_DOWN);
		break;
	case C_FRAME:
		CMP ("frame"); NL;
		CMP ("frame"); EQ; X_I(0); NL;
		NL;
		tk = f0m_tkin_new (tid, C_FRAME, i[0]);
		break;
	case C_INFO_LOCALS:
	case C_INFO_ARGS:
 		CMP ("info_locals_args"); NL;
		NL;
		tk = f0m_tkin_new (tid, code);
		break;
	case C_INFO_LINE:
 		CMP ("info_line"); NL;
		CMP ("file"); EQ; X_S(0); NL;
		CMP ("line"); EQ; X_I(0); NL;
		CMP ("function"); EQ; X_S(1); NL;
		NL;
		tk = f0m_tkin_new (tid, C_INFO_LINE, s[0].c, i[0], s[1].c);
		break;
	case C_INFO_PROGRAM:
 		CMP ("info_program"); NL;
		NL;
		tk = f0m_tkin_new (tid, C_INFO_PROGRAM);
		break;
	case C_INFO_STACK:
 		CMP ("info_stack"); NL;
		NL;
		tk = f0m_tkin_new (tid, C_INFO_STACK);
		break;
	case C_INFO_BREAK:
 		CMP ("info_break"); NL;
		NL;
		tk = f0m_tkin_new (tid, C_INFO_BREAK);
		break;
	case C_INFO_DISPLAY:
 		CMP ("info_display"); NL;
		NL;
		tk = f0m_tkin_new (tid, C_INFO_DISPLAY);
		break;
	case C_THREAD:
		CMP ("thread");
		CMP ("threadID"); EQ; X_I(0); NL;
		NL;
		tk = f0m_tkin_new (tid, C_THREAD, i[0]);
		break;
	case C_INFO_THREADS:
 		CMP ("info_threads"); NL;
		NL;
		tk = f0m_tkin_new (tid, C_INFO_THREADS);
		break;
	case C_SENDTO:
 		CMP ("sendto");
		NL;
		CMP ("command"); EQ; X_S(0); NL;
		NL;
		tk = f0m_tkin_new (tid, C_SENDTO, s[0].c);
		break;
	case C_TERM_SERVER:
 		CMP ("terminate_server"); NL;
		CMP ("halt"); EQ; X_I(0); NL;
		NL;
		tk = f0m_tkin_new (tid, C_TERM_SERVER, i[0]);
		break;
	case C_REGISTER_CLIENT:
 		CMP ("register_client"); NL;
		CMP ("pid"); EQ; X_I(0); NL;
		CMP ("filename"); EQ; X_S(0); NL;
		NL;
		tk = f0m_tkin_new (tid, C_REGISTER_CLIENT, i[0], s[0].c);
		break;
	default:
		msg_pid_fatal ("%s:%d: Invalid code = %d\n",
					   __FILE__, __LINE__, tk->code);
		break;
	}
	NL;
	return tk;
}



/* -----------------------------------------------------------------
 * Function:    tkout_t *f0m_jml_2_tkout (const char *str)
 *
 * Description: Convert a string into a 'tkout_t' strucuture
 *
 * Returns:     TheA dynamically allocated 'tkout_t' strucuture
 */
PUBLIC tkout_t *f0m_jml_2_tkout (const char *str)
{
	
	char *p;
	tkout_t *tk;
	ch_data_t zz;

	char *outp;
	int stat;
	int tid;

	p = (char *)str;
	CMP ("tkout");				/* Check it's a "tkout" */
	NL;							/* Skip line */
	CMP ("tid");				/* Check next field it's "tid" */
	EQ;							/* Skip till number */
	X_II(tid);					/* Extract Task ID form string */
	NL;
	CMP ("str");				/* Check if output is present" */
	EQ;							/* Skip till string */
	X_SS(outp);					/* Get output string */
	NL;							/* Skip line */
	CMP ("status");				/* Check next field it's "code" */
	EQ;							/* Skip till number */
	X_II(stat);					/* Extract status form string */
	NL;
	tk = f0s_tkout_new (stat);
	tk->tid = tid;
	tk->str = outp;
	tk->rep = jml_2_data ("rep", &p);
	zz.code = -1;
	zz.str = &p;
	zz.data = NULL;
	tk->oob = extract_chain (jml_2_oob, "oob", &zz);
	p = *zz.str;
	tk->pos = jml_2_data ("pos", &p);
	tk->err = jml_2_data ("err", &p);
	NL;
	return tk;
}

