/* *****************************************************************
 *
 *  $Id: f0m_rw_to_file.c,v 1.7 2002/09/27 12:43:02 jml Exp $
 *
 *  Copyright (c) 2000 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <03/03/26 23:13:25 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <alloca.h>
#include <assert.h>
#include "f0m_xdr.h"
#include "f0m_rw_to_file.h"



/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */

int readN (int fd, void *buf, int nbytes) {
	int nread, nleft;
	char * bufc = buf;
	
	nleft = nbytes;
	while( nleft > 0) {
		nread = read( fd, bufc, nleft);
		if( nread < 0)              /* Erro */
			return nread;
		if( nread == 0)             /* EOF */
			break;
		nleft -= nread;
		bufc += nread;
	}
	return nbytes - nleft;        /* >= 0 */
}

int writeN (int fd, void *buf, int nbytes) {
	int nwrite, nleft;
	const char * bufc = buf;
	
	nleft = nbytes;
	while( nleft > 0) {
		nwrite = write( fd, bufc, nleft);
		if( nwrite < 0)              /* Erro */
			return nwrite;
		nleft -= nwrite;
		bufc += nwrite;
	}
	return nbytes - nleft;        /* >= 0 */
}


/* #################################################################
 *		PRIVATE functions
 * ################################################################# */



/* -----------------------------------------------------------------
 * Function:    packed_t *pack_tkin (const tkin_t *tk, int *size)
 *
 * Description: Pack a 'tkin_t' strucuture
 *
 * Returns:     The dynamically allocated packed 'tkin_t' strucuture
 */
PRIVATE packed_t *pack_tkin (const tkin_t *tk, int *size)
{
	char *data;
	f0m_init_xdr (XDR_UNDEF);
	data = f0m_tkin_2_xdr (tk);
	if (size != NULL)
		*size = strlen (data);
	return (packed_t *)data;
}



/* -----------------------------------------------------------------
 * Function:    tkout_t *pack_tkout (const tkout_t *tk, int *size)
 *
 * Description: Pack a 'tkout_t' strucuture
 *
 * Returns:     The dynamically allocated packed 'tkout_t' strucuture
 */
PRIVATE packed_t *pack_tkout (const tkout_t *tk, int *size)
{
	char *data;
	f0m_init_xdr (XDR_UNDEF);
	data = f0m_tkout_2_xdr (tk);
	if (size != NULL)
		*size = strlen (data);
	return (packed_t *)data;
}



/* -----------------------------------------------------------------
 * Function:    tkin_t *unpack_tkin (packed_t *data)
 *
 * Description: Unpack a 'tkin_t' strucuture
 *
 * Returns:     The dynamically allocated unpacked 'tkin_t' strucuture
 */
tkin_t *unpack_tkin (packed_t *data)
{
	f0m_init_xdr (XDR_UNDEF);
	return f0m_xdr_2_tkin ((char *)data);
}



/* -----------------------------------------------------------------
 * Function:    tkout_t *unpack_tkout (packed_t *data)
 *
 * Description: Unpack a 'tkout_t' strucuture
 *
 * Returns:     The dynamically allocated unpacked 'tkout_t' strucuture
 */
tkout_t *unpack_tkout (packed_t *data)
{
	f0m_init_xdr (XDR_UNDEF);
	return f0m_xdr_2_tkout ((char *)data);
}


/* -----------------------------------------------------------------
 * Function:    void f0m_log_communications()
 *
 * Description: Check environment variable 'FIDDLE_TK_FILE'. If set,
 *				it should contain the name of the file to log the tokens
 *
 * Returns:     []
 */
typedef enum {TKIN, TKOUT} tk_type_t;
void f0m_log_communications (tk_type_t tktype, const char *tk)
{
	static int tested_f0m_tk_file = 0;
	/* =  0		not tested yet */
	/* = -1		tested and not active */
	/* =  1		tested and active */
	static const char *f0m_tk_file;
	static FILE *log;

	if (tested_f0m_tk_file == 0) {
		/* Not tested yet */
		f0m_tk_file = getenv ("FIDDLE_TK_FILE");
		tested_f0m_tk_file = (f0m_tk_file == NULL) ? -1 : 1;
		if (tested_f0m_tk_file > 0) {
			log = fopen (f0m_tk_file, "w");
			if (log == NULL)
				msg_pid_fatal ("Couldn't open \"%s\" for writing\n\n",
							   f0m_tk_file);
			fclose (log);
		}
	}
	if (tested_f0m_tk_file > 0) {
		/* Tested and active */
		log = fopen (f0m_tk_file, "a");
		fprintf (log, "\n===========================================================================\n");
		switch (tktype) {
		case TKIN:
			fprintf (log, "== TKIN == TKIN == TKIN == TKIN == TKIN == TKIN == TKIN == TKIN == TKIN ===\n");
			break;
		case TKOUT:
			fprintf (log, "== TKOUT == TKOUT == TKOUT == TKOUT == TKOUT == TKOUT == TKOUT == TKOUT ===\n");
			break;
		default:
			msg_pid_fatal ("Invalid valur for tk_type");
		}
		fprintf (log, "===========================================================================\n");
		fprintf (log, "%s", tk);
		fprintf (log, "===========================================================================\n");
		fclose(log);
	}
}


/* #################################################################
 *		PUBLIC functions
 * ################################################################# */


/* -----------------------------------------------------------------
 * Function:    int status = f0m_write_tkin (int fd, const tkin_t *tk)
 *
 * Description: Write a 'tkin_t' strucuture to a file in a machine
 *				independent way
 *
 * Returns:     The same as 'write()'
 */
PUBLIC int f0m_write_tkin (int fd, const tkin_t *tk)
{
	int nb;
	int bytes_to_write;
	packed_t *ptk;

	ptk = pack_tkin (tk, &bytes_to_write);
	assert (ptk != NULL);
	nb = writeN (fd, &bytes_to_write, sizeof (bytes_to_write));
	assert (nb == sizeof (bytes_to_write));
	nb = writeN (fd, ptk, bytes_to_write);
	assert (nb == bytes_to_write);
	fdatasync (fd);
	f0m_log_communications (TKIN, (char *)ptk);
	free (ptk);
	return nb;
}


/* -----------------------------------------------------------------
 * Function:    int status = f0m_write_tkout (int fd, const tkout_t *tk)
 *
 * Description: Write a 'tkout_t' strucuture to a file in a machine
 *				independent way
 *
 * Returns:     The same as 'write()'
 */
PUBLIC int f0m_write_tkout (int fd, const tkout_t *tk)
{
	int nb;
	int bytes_to_write;
	packed_t *ptk;

	ptk = pack_tkout (tk, &bytes_to_write);
	assert (ptk != NULL);
	/*	usleep(100);				/ * BUG! BUG! BUG! BUG! BUG! BUG! BUG! */
	nb = writeN (fd, &bytes_to_write, sizeof (bytes_to_write));
	assert (nb == sizeof (bytes_to_write));
	nb = writeN (fd, ptk, bytes_to_write);
	assert (nb == bytes_to_write);
	fdatasync (fd);
	free (ptk);
	return nb;
}


/* -----------------------------------------------------------------
 * Function:    tkin_t *tk = f0m_read_tkin (int fd)
 *
 * Description: Read a 'tkin_t' strucuture from a file
 *
 * Returns:     The same as 'read()'
 */
PUBLIC tkin_t *f0m_read_tkin (int fd)
{
	int nb;
	int bytes_to_read;
	packed_t *data;

	nb = readN (fd, &bytes_to_read, sizeof (bytes_to_read));
	if (nb == 0) return NULL;	/* EOF */
	assert (nb == sizeof (bytes_to_read));
	data = alloca (bytes_to_read + 1);
	assert (data != NULL);
	nb = readN (fd, data, bytes_to_read);
	assert (nb = bytes_to_read);
	((char *)data)[bytes_to_read] = '\0';
	return unpack_tkin (data);
}


/* -----------------------------------------------------------------
 * Function:    tkout_t *tk = f0m_read_tkout (int fd)
 *
 * Description: Read a 'tkout_t' strucuture from a file
 *
 * Returns:     The same as 'read()'
 */
PUBLIC tkout_t *f0m_read_tkout (int fd)
{
	int nb;
	int bytes_to_read;
	packed_t *data;

	nb = readN (fd, &bytes_to_read, sizeof (bytes_to_read));
	if (nb == 0) return NULL;
	assert (nb == sizeof (bytes_to_read));
	data = alloca (bytes_to_read + 1);
	assert (data != NULL);
	nb = readN (fd, data, bytes_to_read);
	if (nb == 0) return NULL;
	assert (nb = bytes_to_read);
	((char *)data)[nb] = '\0';
	f0m_log_communications (TKOUT, (const char *)data);
	return unpack_tkout (data);
}

