/* *****************************************************************
 *
 *  $Id: f0m_x_2_jml.c,v 1.1 2002/09/27 12:43:02 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <03/03/26 23:15:40 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_ALLOCO_H
#include <alloca.h>
#endif
#include <assert.h>

#include "glob.h"
#include "chain.h"
#include "msg.h"

#include "f0s_tkout.h"

#include "f0m_tkin.h"
#include "f0m_cmd_codes.h"
#include "f0m_x_2_jml.h"


typedef struct mychar_s mychar_t;
struct mychar_s {
	char *str;
	int len;
	int max;
};






PRIVATE mychar_t *mychar_append (mychar_t *s, const char *str)
{
	int l = strlen (str);
	while (s->len + l >= s->max) {
		/* More space needed */
		s->max *=2;
		s->str = (char *)realloc (s->str, s->max);
		assert (s->str != NULL);
	}
 	memcpy (s->str + s->len, str, l+1);
	s->len += l;
	return s;
}

PRIVATE char *mychar_thestr (mychar_t *s)
{
	char *ss = s->str;
	free (s);
	return ss;
}

PRIVATE mychar_t *mychar_new (void)
{
	mychar_t *s = calloc (1, sizeof (*s));
	assert (s != NULL);
	s->len = 0;
	s->max = 8;
	s->str = calloc (s->max, sizeof (char));
	return s;
}

/*
PRIVATE mychar_t *mychar_new_s (const char *str)
{
	mychar_t *s = mychar_new ();
	return mychar_append (s, str);
}

PRIVATE void mychar_delete (mychar_t *s)
{
	free (s->str);
	free (s);
}
*/



/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */

#define MAX_BUF 64*1024

#define STRLEN(s)		(s) == NULL ? 0 : strlen (s)
#define STRING(s)		(s) == NULL ? "" : (s)

#define N_IND "  "


#define INDENT(s,n)										\
		do {											\
			int _xx = (n);								\
			while (_xx--)								\
					mychar_append ((s), N_IND);			\
		} while (0)

#define sprinta2(s,n,f,x,y)								\
		do {											\
			char buf[MAX_BUF];							\
			sprintf (buf, (f), (x), (y));				\
			INDENT((s), (n));							\
			mychar_append ((s), buf);					\
		} while (0)
	
#define sprinta1(s,n,f,x)								\
		do {											\
			char buf[MAX_BUF];							\
			sprintf (buf, (f), (x));					\
			INDENT((s), (n));							\
			mychar_append ((s), buf);					\
		} while (0)

#define sprinta0(s,n,f)									\
		do {											\
			char buf[MAX_BUF];							\
			sprintf (buf, (f));							\
			INDENT((s), (n));							\
			mychar_append ((s), buf);					\
		} while (0)


typedef struct sprint_data_s sprint_data_t;
struct sprint_data_s {
	int			level;
	code_t		code;
	mychar_t   *str;
};




/* #################################################################
 *		PRIVATE functions
 * ################################################################# */


PRIVATE void sprinta_chain (chain_t *chain, chain_walker_f walker,
							const char *name, sprint_data_t *spd)
{
	sprinta2 (spd->str, spd->level, "chain[%d]:%s = {\n",
			  chain_length (chain), name);
	spd->level++;
	chain_walk (chain, walker, spd);
	spd->level--;
	sprinta0 (spd->str, spd->level, "}\n");
}



PRIVATE int sprinta_chain_node (void *obj, void *args)
{
	sprint_data_t *spd = (sprint_data_t *)args;
	switch (spd->code) {
	case R_TIDS:
	{   rtids_t *d = (rtids_t *)obj;
		sprinta0 (spd->str, spd->level, "tid = {\n");
		sprinta1 (spd->str, spd->level+1, "gtid = %d\n", d->gtid);
		sprinta1 (spd->str, spd->level+1, "tp_pid = %d\n", d->tp_pid);
		sprinta1 (spd->str, spd->level+1, "dbg_pid = %d\n", d->dbg_pid);
		sprinta1 (spd->str, spd->level+1, "attached = %d\n", d->attached);
		sprinta1 (spd->str, spd->level+1, "ltid = %d\n", d->ltid);
		sprinta2 (spd->str, spd->level+1, "machine = %d %s\n",
				  STRLEN (d->machine), STRING (d->machine));
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_CLIENTS:
	{   rclients_t *d = (rclients_t *)obj;
		sprinta0 (spd->str, spd->level, "clients = {\n");
		sprinta1 (spd->str, spd->level+1, "pid = %d\n", d->pid);
		sprinta2 (spd->str, spd->level+1, "filename = %d %s\n",
				  STRLEN (d->filename), STRING (d->filename));
		sprinta2 (spd->str, spd->level+1, "hostaddr = %d %s\n",
				  STRLEN (d->hostaddr), STRING (d->hostaddr));
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_ADD_DEBUGGER:
	{   radddebugger_t *d = (radddebugger_t *)obj;
	    sprinta0 (spd->str, spd->level, "add_debugger = {\n");
		sprinta1 (spd->str, spd->level+1, "tid = %d\n", d->tid);
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_KILL:
	{   sprinta0 (spd->str, spd->level, "kill = {\n");
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_ATTACH:
	{   rattach_t *d = (rattach_t *)obj;
		sprinta0 (spd->str, spd->level, "attach = {\n");
		sprinta1 (spd->str, spd->level+1, "tid = %d\n", d->tid);
		sprinta1 (spd->str, spd->level+1, "tp_pid = %d\n", d->tp_pid);
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_DETACH:
	{   rdetach_t *d = (rdetach_t *)obj;
		sprinta0 (spd->str, spd->level, "detach = {\n");
		sprinta1 (spd->str, spd->level+1, "tp_pid = %d\n", d->tp_pid);
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_SYMBOL_FILE:
	{   /* rsymbolfile_t *d = (rsymbolfile_t *)obj; */
		sprinta0 (spd->str, spd->level, "symbol_file = {\n");
/* 		sprinta2 (spd->str, spd->level+1, "file = %d %s\n", */
/* 				  STRLEN (d->file), */
/* 				  STRING (d->file)); */
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_FILE:
	{   rfile_t *d = (rfile_t *)obj;
		sprinta0 (spd->str, spd->level, "file = {\n");
		sprinta1 (spd->str, spd->level+1, "tid = %d\n", d->tid);
/* 		sprinta2 (spd->str, spd->level+1, "file = %d %s\n", */
/* 				  STRLEN (d->file), */
/* 				  STRING (d->file)); */
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_RUN:
	{   rrun_t *d = (rrun_t *)obj;
		sprinta0 (spd->str, spd->level, "run = {\n");
		sprinta2 (spd->str, spd->level+1, "file = %d %s\n",
				  STRLEN (d->file),
				  STRING (d->file));
		sprinta0 (spd->str, spd->level, "}\n");
   		break;
	}
	case R_STEP:
	{   /* rstep_t *d = (rstep_t *)obj; */
		sprinta0 (spd->str, spd->level, "step = {\n");
		/* Empty */
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_NEXT:
	{   /* rnext_t *d = (rnext_t *)obj; */
		sprinta0 (spd->str, spd->level, "next = {\n");
   		/* Empty */
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_CONTINUE:
	{   /* rcontinue_t *d = (rcontinue_t *)obj; */
		sprinta0 (spd->str, spd->level, "continue = {\n");
		/* Empty */
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_FINISH:
	{   rfinish_t *d = (rfinish_t *)obj;
		sprinta0 (spd->str, spd->level, "finish = {\n");
		sprinta1 (spd->str, spd->level+1, "s_address = 0x%x\n",
				  (unsigned)(d->s_address));
		sprinta2 (spd->str, spd->level+1, "s_function = %d %s\n",
				  STRLEN (d->s_function),
				  STRING (d->s_function));
		sprinta2 (spd->str, spd->level+1, "s_args = %d %s\n",
				  STRLEN (d->s_args),
				  STRING (d->s_args));
		sprinta2 (spd->str, spd->level+1, "s_file = %d %s\n",
				  STRLEN (d->s_file),
				  STRING (d->s_file));
		sprinta1 (spd->str, spd->level+1, "s_line = %d\n",
				  d->s_line);
		sprinta1 (spd->str, spd->level+1, "dbg_var = %d\n",
				  d->dbg_var);
		sprinta2 (spd->str, spd->level+1, "value = %d %s\n",
				  STRLEN (d->value),
				  STRING (d->value));
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_CALL_VOID:
	{   /* rcall_t *d; * = (rcall_t *)obj */
		sprinta0 (spd->str, spd->level, "call_void = {\n");
		/* Empty */
   		sprinta0 (spd->str, spd->level, "}\n");
   		break;
	}
	case R_BREAK:
	{   rbreak_t *d = (rbreak_t *)obj;
		sprinta0 (spd->str, spd->level, "break = {\n");
		sprinta1 (spd->str, spd->level+1, "bpid = %d\n", d->bpid);
		sprinta1 (spd->str, spd->level+1, "address = 0x%x\n",
				  (unsigned)(d->address));
		sprinta2 (spd->str, spd->level+1, "file = %d %s\n",
				  STRLEN (d->file),
				  STRING (d->file));
		sprinta1 (spd->str, spd->level+1, "line = %d\n", d->line);
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_DELETE:
	{   /* rdelete_t *d = (rdelete_t *)obj; */
		sprinta0 (spd->str, spd->level, "delete = {\n");
		/* Empty */
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_SET_VARIABLE:
	{   /* rsetvariable_t *d = (rsetvariable_t *)obj; */
		sprinta0 (spd->str, spd->level, "set_variable = {\n");
		/* Empty */
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_EVALUATE:
	{   revaluate_t *d = (revaluate_t *)obj;
		sprinta0 (spd->str, spd->level, "evaluate = {\n");
		sprinta1 (spd->str, spd->level+1, "dbg_var = %d\n",
				  d->dbg_var);
		sprinta2 (spd->str, spd->level+1, "value = %d %s\n",
				  STRLEN (d->value),
				  STRING (d->value));
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_DISPLAY_EXPRESSION:
	{   /* rdisplay_t *d = (rdisplay_t *)obj; */
		sprinta0 (spd->str, spd->level, "display_expression = {\n");
		/* 		sprinta_chain (d->data, sprint_display, str, level+1); */
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_UNDISPLAY:
	{   sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_LIST:
	{   rlist_t *d = (rlist_t *)obj;
		sprinta0 (spd->str, spd->level, "list = {\n");
		sprinta1 (spd->str, spd->level+1, "line_number = %d\n", d->line_number);
		sprinta2 (spd->str, spd->level+1, "line_text = %d %s\n",
				  STRLEN (d->line_text), STRING (d->line_text));
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_INFO_LOCALS:
	{   rinfolocals_t *d = (rinfolocals_t *)obj;
		sprinta0 (spd->str, spd->level, "info_locals_args = {\n");
		sprinta2 (spd->str, spd->level+1, "var = %d %s\n",
				  STRLEN (d->var), STRING (d->var));
		sprinta2 (spd->str, spd->level+1, "value = %d %s\n",
				  STRLEN (d->value), STRING (d->value));
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_INFO_LINE_NO_CODE:
	case R_INFO_LINE:
	{   rinfoline_t *d = (rinfoline_t *)obj;
		sprinta0 (spd->str, spd->level, "info_line = {\n");
		sprinta1 (spd->str, spd->level+1, "line = %d\n", d->line);
		sprinta2 (spd->str, spd->level+1, "file = %d %s\n",
				  STRLEN (d->file),
				  STRING (d->file));
		sprinta2 (spd->str, spd->level+1, "start_function = %d %s\n",
				  STRLEN (d->start_function),
				  STRING (d->start_function));
		sprinta1 (spd->str, spd->level+1, "start_function_offset = %d\n",
				  d->start_function_offset);
		sprinta2 (spd->str, spd->level+1, "end_function = %d %s\n",
				  STRLEN (d->end_function),
				  STRING (d->end_function));
		sprinta1 (spd->str, spd->level+1, "end_function_offset = %d\n",
				  d->end_function_offset);
		sprinta1 (spd->str, spd->level+1, "start_address = 0x%x\n",
				  (unsigned)(d->start_address));
		sprinta1 (spd->str, spd->level+1, "end_address = 0x%x\n",
				  (unsigned)(d->end_address));
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_INFO_PROGRAM:
	{   rinfoprogram_t *d = (rinfoprogram_t *)obj;
		sprinta0 (spd->str, spd->level, "info_program = {\n");
		sprinta1 (spd->str, spd->level+1, "tp_pid = %d\n",
				  d->tp_pid);
		sprinta1 (spd->str, spd->level+1, "address = 0x%x\n",
				  (unsigned)(d->address));
		sprinta2 (spd->str, spd->level+1, "why = %d %s\n",
				  STRLEN (d->why),
				  STRING (d->why));
		sprinta1 (spd->str, spd->level+1, "bpid = %d\n", d->bpid);
		sprinta2 (spd->str, spd->level+1, "signal = %d %s\n",
				  STRLEN (d->signal),
				  STRING (d->signal));
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_INFO_STACK:
	{   rframe_t *d = (rframe_t *)obj;
		sprinta0 (spd->str, spd->level, "info_stack = {\n");
		sprinta1 (spd->str, spd->level+1, "frame = %d\n", d->frame);
		sprinta1 (spd->str, spd->level+1, "address = 0x%x\n",
				  (unsigned)(d->address));
		sprinta2 (spd->str, spd->level+1, "function = %d %s\n",
				  STRLEN (d->function), STRING (d->function));
		sprinta2 (spd->str, spd->level+1, "arguments = %d %s\n",
				  STRLEN (d->arguments), STRING (d->arguments));
		sprinta2 (spd->str, spd->level+1, "file = %d %s\n",
				  STRLEN (d->file), STRING (d->file));
		sprinta1 (spd->str, spd->level+1, "line = %d\n", d->line);
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_INFO_BREAK:
	{   rinfobreak_t *d = (rinfobreak_t *)obj;
		sprinta0 (spd->str, spd->level, "info_break = {\n");
		sprinta1 (spd->str, spd->level+1, "bpid = %d\n", d->bpid);
		sprinta2 (spd->str, spd->level+1, "type = %d %s\n",
				  STRLEN (d->type), STRING (d->type));
		sprinta2 (spd->str, spd->level+1, "disp = %d %s\n",
				  STRLEN (d->disp), STRING (d->disp));
		sprinta1 (spd->str, spd->level+1, "enabled = %d\n", d->enabled);
		sprinta1 (spd->str, spd->level+1, "address = 0x%x\n",
				  (unsigned)(d->address));
		sprinta2 (spd->str, spd->level+1, "function = %d %s\n",
				  STRLEN (d->function), STRING (d->function));
		sprinta2 (spd->str, spd->level+1, "file = %d %s\n",
				  STRLEN (d->file), STRING (d->file));
		sprinta1 (spd->str, spd->level+1, "line = %d\n", d->line);
		sprinta1 (spd->str, spd->level+1, "hit_counter = %d\n", d->hit_counter);
		sprinta1 (spd->str, spd->level+1, "ign_counter = %d\n", d->ign_counter);
		sprinta2 (spd->str, spd->level+1, "condition = %d %s\n",
				  STRLEN (d->condition), STRING (d->condition));
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_INFO_DISPLAY:
	{   rdisplay_t *d = (rdisplay_t *)obj;
		sprinta0 (spd->str, spd->level, "info_display = {\n");
		sprinta1 (spd->str, spd->level+1, "dispid = %d\n", d->dispid);
		sprinta1 (spd->str, spd->level+1, "enabled = %d\n", d->enabled);
		sprinta2 (spd->str, spd->level+1, "expression = %d %s\n",
				  STRLEN (d->expression), STRING (d->expression));
		sprinta2 (spd->str, spd->level+1, "value = %d %s\n",
				  STRLEN (d->value), STRING (d->value));
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_INFO_THREADS:
	{   rinfothreads_t *d = (rinfothreads_t *)obj;
		sprinta0 (spd->str, spd->level, "info_threads = {\n");
		sprinta1 (spd->str, spd->level+1, "active = %d\n", d->active);
		sprinta1 (spd->str, spd->level+1, "th_id = %d\n", d->th_id);
		sprinta1 (spd->str, spd->level+1, "tp_pid = %d\n", d->tp_pid);
		sprinta1 (spd->str, spd->level+1, "frame = 0x%x\n",
				  (unsigned)(d->frame));
		sprinta2 (spd->str, spd->level+1, "function = %d %s\n",
				  STRLEN (d->function), STRING (d->function));
		sprinta2 (spd->str, spd->level+1, "arguments = %d %s\n",
				  STRLEN (d->arguments), STRING (d->arguments));
		sprinta2 (spd->str, spd->level+1, "file = %d %s\n",
				  STRLEN (d->file), STRING (d->file));
		sprinta1 (spd->str, spd->level+1, "line = %d\n", d->line);
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_TTY:
	{   /* rtty_t *d = (rtty_t *)obj; */
		sprinta0 (spd->str, spd->level, "tty = {\n");
		/* Empty */
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_THREAD:
	{   /* rthread_t *d; */
		sprinta0 (spd->str, spd->level, "thread = {\n");
		/* Empty */
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case R_SENDTO:
	{   /* rsendto_t *d; */
		sprinta0 (spd->str, spd->level, "sendto = {\n");
		/* Empty -- output is already on 'rep->str' */
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ */
	case O_DISPLAY:
	{   odisplay_t *d = (odisplay_t *)obj;
		sprinta0 (spd->str, spd->level, "display = {\n");
		sprinta1 (spd->str, spd->level+1, "dispid = %d\n", d->dispid);
		sprinta1 (spd->str, spd->level+1, "enabled = %d\n", d->enabled);
		sprinta2 (spd->str, spd->level+1, "expression = %d %s\n",
				  STRLEN (d->expression), STRING (d->expression));
		sprinta2 (spd->str, spd->level+1, "value = %d %s\n",
				  STRLEN (d->value), STRING (d->value));
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case O_BPHIT:
	{   obphit_t *d = (obphit_t *)obj;
		sprinta0 (spd->str, spd->level, "bp_hit = {\n");
		sprinta1 (spd->str, spd->level+1, "bpid = %d\n", d->bpid);
		sprinta2 (spd->str, spd->level+1, "function = %d %s\n",
				  STRLEN (d->function),
				  STRING (d->function));
		sprinta2 (spd->str, spd->level+1, "args = %d %s\n",
				  STRLEN (d->args),
				  STRING (d->args));
		sprinta2 (spd->str, spd->level+1, "file = %d %s\n",
				  STRLEN (d->file),
				  STRING (d->file));
		sprinta1 (spd->str, spd->level+1, "line = %d\n", d->line);
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case O_SIGNAL:
	{   osignal_t *d = (osignal_t *)obj;
		sprinta0 (spd->str, spd->level, "signal = {\n");
		sprinta2 (spd->str, spd->level+1, "name = %d %s\n",
				  STRLEN (d->name),
				  STRING (d->name));
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case O_NEW_THREAD:
	{   onewthread_t *d = (onewthread_t *)obj;
		sprinta0 (spd->str, spd->level, "new_thread = {\n");
		sprinta1 (spd->str, spd->level+1, "id = %d\n", d->id);
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	case O_SWITCH_TO_THREAD:
	{   othread_t *d = (othread_t *)obj;
		sprinta0 (spd->str, spd->level, "thread = {\n");
		sprinta1 (spd->str, spd->level+1, "pid = %d\n", d->pid);
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@ */
	case P_POSITION:
	{   pposition_t *d = (pposition_t *)obj;
		sprinta0 (spd->str, spd->level, "position = {\n");
		sprinta2 (spd->str, spd->level+1, "file = %d %s\n",
				  STRLEN (d->file),
				  STRING (d->file));
		sprinta1 (spd->str, spd->level+1, "line = %d\n",
				  d->line);
		sprinta1 (spd->str, spd->level+1, "pos = %d\n",
				  d->pos);
		sprinta2 (spd->str, spd->level+1, "where = %d %s\n",
				  STRLEN (d->where),
				  STRING (d->where));
		sprinta1 (spd->str, spd->level+1, "frame = 0x%x\n",
				  (unsigned)(d->frame));
		sprinta0 (spd->str, spd->level, "}\n");
		break;
	}
	default:
		assert (0);
	}
	return 1;
}





/* -----------------------------------------------------------------
 * Function:	void oob_2_jml (const oobdata_t *oob, mychar_t *str, int level)
 *
 * Description: Convert a 'oobdata_t' to text
 *
 * Returns:     []
 */
PRIVATE  int oob_2_jml  (void *obj, void *args)
{
	data_t *oob = (data_t *)obj;
	sprint_data_t *spd = (sprint_data_t *)args;

	sprinta0 (spd->str, spd->level, "oob = {\n");
	if (oob != NULL) {
		spd->level++;					/* Increment indentation level */
		sprinta2 (spd->str, spd->level, "str = %d %s\n",
				  STRLEN (oob->str), STRING (oob->str));
		sprinta1 (spd->str, spd->level, "code = %d\n", oob->code);
		spd->code = oob->code;
		sprinta_chain (oob->data, sprinta_chain_node, "data", spd);
	}
	sprinta0 (spd->str, --spd->level, "}\n");
	return 1;
}




/* -----------------------------------------------------------------
 * Function:	void data_2_jml (const char *varname, const data_t *var,
 *								 const mychar *str, int level)
 *
 * Description: Convert a 'data_t' to text
 *
 * Returns:     []
 */
PRIVATE  void data_2_jml (const char *varname, const data_t *var,
						  mychar_t *str, int level)
{
	sprint_data_t spd;

	sprinta1 (str, level, "%s = {\n", varname);
	if (var != NULL) {
		spd.code = var->code;
		spd.level = level+1;
		spd.str = str;
   		sprinta2 (spd.str, spd.level, "str = %d %s\n",
				  STRLEN (var->str), STRING (var->str));
		sprinta1 (spd.str, spd.level, "code = %d\n", var->code);
		
		sprinta_chain (var->data, sprinta_chain_node, "data", &spd);
	}
	sprinta0 (str, level, "}\n");
}



/* #################################################################
 *		PUBLIC functions
 * ################################################################# */

/* -----------------------------------------------------------------
 * Function: char *f0m_tkin_2_jml (const tkin_t *tk)
 *
 * Description: Convert a 'tkin_t' to text
 *
 * Returns:     A dynamically allocated buffer with the text
 */
char *f0m_tkin_2_jml (const tkin_t *tk)
{
	int level;
	mychar_t *str;

	str = mychar_new ();
	sprinta0 (str, 0, "tkin = {\n");
	level = 1;					/* Indentation = 1 */
	sprinta1 (str, level, "tid = %d\n", tk->tid);
	sprinta1 (str, level, "code = %d\n", tk->code);
 	switch (tk->code) {
 	case C_TIDS:
 		sprinta0 (str, level, "tids = {\n");
		sprinta2 (str, level+1, "options = %d %s\n",
				  STRLEN (tk->data.r_tids_clients.options),
				  STRING (tk->data.r_tids_clients.options));
		sprinta0 (str, level, "}\n");
		break;
 	case C_CLIENTS:
 		sprinta0 (str, level, "clients = {\n");
		sprinta2 (str, level+1, "options = %d %s\n",
				  STRLEN (tk->data.r_tids_clients.options),
				  STRING (tk->data.r_tids_clients.options));
		sprinta0 (str, level, "}\n");
		break;
	case C_KILL:
 		sprinta0 (str, level, "kill = {\n");
		sprinta0 (str, level, "}\n");
		break;
	case C_ADD_DEBUGGER:
 		sprinta0 (str, level, "add_debugger = {\n");
		sprinta2 (str, level+1, "host = %d %s\n",
				  STRLEN (tk->data.r_add_debugger.host),
				  STRING (tk->data.r_add_debugger.host));
		sprinta1 (str, level+1, "dbg_pid = %d\n", tk->data.r_add_debugger.dbg_pid);
		sprinta1 (str, level+1, "tp_pid = %d\n", tk->data.r_add_debugger.tp_pid);
		sprinta2 (str, level+1, "cmd_channel = %d %s\n",
				  STRLEN (tk->data.r_add_debugger.cmd_channel),
				  STRING (tk->data.r_add_debugger.cmd_channel));
		sprinta2 (str, level+1, "rep_channel = %d %s\n",
				  STRLEN (tk->data.r_add_debugger.rep_channel),
				  STRING (tk->data.r_add_debugger.rep_channel));
		sprinta0 (str, level, "}\n");
		break;
	case C_ATTACH:
 		sprinta0 (str, level, "attach = {\n");
		sprinta1 (str, level+1, "tp_pid = %d\n", tk->data.r_attach.tp_pid);
		sprinta2 (str, level+1, "host = %d %s\n",
				  STRLEN (tk->data.r_attach.host),
				  STRING (tk->data.r_attach.host));
		sprinta0 (str, level, "}\n");
		break;
	case C_DETACH:
 		sprinta0 (str, level, "detach = {\n");
		sprinta0 (str, level, "}\n");
		break;
	case C_SYMBOL_FILE:
 		sprinta0 (str, level, "symbol_file = {\n");
		sprinta2 (str, level+1, "file = %d %s\n",
				  STRLEN (tk->data.r_symbol_file.file),
				  STRING (tk->data.r_symbol_file.file));
		sprinta0 (str, level, "}\n");
		break;
	case C_FILE:
 		sprinta0 (str, level, "file = {\n");
		sprinta2 (str, level+1, "file = %d %s\n",
				  STRLEN (tk->data.r_file.file),
				  STRING (tk->data.r_file.file));
		sprinta2 (str, level+1, "host = %d %s\n",
				  STRLEN (tk->data.r_file.host),
				  STRING (tk->data.r_file.host));
		sprinta0 (str, level, "}\n");
		break;
	case C_RUN:
 		sprinta0 (str, level, "run = {\n");
		sprinta2 (str, level+1, "args = %d %s\n",
				  STRLEN (tk->data.r_run.args),
				  STRING (tk->data.r_run.args));
		sprinta0 (str, level, "}\n");
		break;
	case C_STEP:
 		sprinta0 (str, level, "step = {\n");
		sprinta0 (str, level, "}\n");
		break;
	case C_NEXT:
 		sprinta0 (str, level, "next = {\n");
		sprinta0 (str, level, "}\n");
		break;
	case C_CONTINUE:
 		sprinta0 (str, level, "continue = {\n");
		sprinta0 (str, level, "}\n");
		break;
	case C_FINISH:
 		sprinta0 (str, level, "finish = {\n");
		sprinta0 (str, level, "}\n");
		break;
	case C_SIGNAL:
		sprinta0 (str, level, "signal = {\n");
		sprinta1 (str, level+1, "tid = %d\n", tk->data.r_signal.tid);
		sprinta1 (str, level+1, "signum = %d\n", tk->data.r_signal.signum);
		sprinta0 (str, level, "}\n");
		break;
	case C_BREAK:
 		sprinta0 (str, level, "break = {\n");
		sprinta2 (str, level+1, "file = %d %s\n",
				  STRLEN (tk->data.r_break.file),
				  STRING (tk->data.r_break.file));
		sprinta1 (str, level+1, "line = %d\n", tk->data.r_break.line);
		sprinta2 (str, level+1, "function = %d %s\n",
				  STRLEN (tk->data.r_break.function),
				  STRING (tk->data.r_break.function));
		sprinta0 (str, level, "}\n");
		break;
	case C_DELETE:
 		sprinta0 (str, level, "delete = {\n");
		sprinta1 (str, level+1, "bpid = %d\n", tk->data.r_delete.bpid);
		sprinta0 (str, level, "}\n");
		break;
	case C_SET_VARIABLE:
 		sprinta0 (str, level, "set_variable = {\n");
		sprinta2 (str, level+1, "variable = %d %s\n",
				  STRLEN (tk->data.r_set_variable.variable),
				  STRING (tk->data.r_set_variable.variable));
		sprinta2 (str, level+1, "expression = %d %s\n",
				  STRLEN (tk->data.r_set_variable.expression),
				  STRING (tk->data.r_set_variable.expression));
		sprinta0 (str, level, "}\n");
		break;
	case C_EVALUATE:
 		sprinta0 (str, level, "evaluate = {\n");
		sprinta2 (str, level+1, "expression = %d %s\n",
				  STRLEN (tk->data.r_evaluate.expression),
				  STRING (tk->data.r_evaluate.expression));
		sprinta0 (str, level, "}\n");
		break;
	case C_DISPLAY:
 		sprinta0 (str, level, "display = {\n");
		sprinta2 (str, level+1, "expression = %d %s\n",
				  STRLEN (tk->data.r_display.expression),
				  STRING (tk->data.r_display.expression));
		sprinta0 (str, level, "}\n");
		break;
	case C_UNDISPLAY:
		sprinta0 (str, level, "undisplay = {\n");
		sprinta1 (str, level+1, "dispid = %d\n", tk->data.r_undisplay.dispid);
		sprinta0 (str, level, "}\n");
		break;
	case C_CALL:
 		sprinta0 (str, level, "call = {\n");
		sprinta2 (str, level+1, "expression = %d %s\n",
				  STRLEN (tk->data.r_call.expression),
				  STRING (tk->data.r_call.expression));
		sprinta0 (str, level, "}\n");
		break;
	case C_LIST:
 		sprinta0 (str, level, "list = {\n");
		sprinta2 (str, level+1, "file = %d %s\n",
				  STRLEN (tk->data.r_list.file),
				  STRING (tk->data.r_list.file));
		sprinta1 (str, level+1, "start_line = %d\n", tk->data.r_list.start_line);
		sprinta2 (str, level+1, "function = %d %s\n",
				  STRLEN (tk->data.r_list.function),
				  STRING (tk->data.r_list.function));
		sprinta1 (str, level+1, "n_lines = %d\n", tk->data.r_list.n_lines);
		sprinta0 (str, level, "}\n");
		break;
	case C_TTY:
 		sprinta0 (str, level, "tty = {\n");
		sprinta2 (str, level+1, "file = %d %s\n",
				  STRLEN (tk->data.r_tty.file),
				  STRING (tk->data.r_tty.file));
		sprinta0 (str, level, "}\n");
		break;
	case C_UP:
 		sprinta0 (str, level, "up = {\n");
		sprinta0 (str, level, "}\n");
		break;
	case C_DOWN:
 		sprinta0 (str, level, "down = {\n");
		sprinta0 (str, level, "}\n");
		break;
	case C_FRAME:
		sprinta0 (str, level, "frame = {\n");
		sprinta1 (str, level+1, "frame = %d\n", tk->data.r_frame.frame);
		sprinta0 (str, level, "}\n");
		break;
	case C_INFO_LOCALS:
	case C_INFO_ARGS:
 		sprinta0 (str, level, "info_locals_args = {\n");
		sprinta0 (str, level, "}\n");
		break;
	case C_INFO_LINE:
 		sprinta0 (str, level, "info_line = {\n");
		sprinta2 (str, level+1, "file = %d %s\n",
				  STRLEN (tk->data.r_info_line.file),
				  STRING (tk->data.r_info_line.file));
		sprinta1 (str, level+1, "line = %d\n", tk->data.r_info_line.line);
		sprinta2 (str, level+1, "function = %d %s\n",
				  STRLEN (tk->data.r_info_line.function),
				  STRING (tk->data.r_info_line.function));
		sprinta0 (str, level, "}\n");
		break;
	case C_INFO_PROGRAM:
 		sprinta0 (str, level, "info_program = {\n");
		sprinta0 (str, level, "}\n");
		break;
	case C_INFO_STACK:
 		sprinta0 (str, level, "info_stack = {\n");
		sprinta0 (str, level, "}\n");
		break;
	case C_INFO_BREAK:
 		sprinta0 (str, level, "info_break = {\n");
		sprinta0 (str, level, "}\n");
		break;
	case C_INFO_DISPLAY:
 		sprinta0 (str, level, "info_display = {\n");
		sprinta0 (str, level, "}\n");
		break;
	case C_THREAD:
		sprinta0 (str, level, "thread = {\n");
		sprinta1 (str, level+1, "threadID = %d\n",
				  tk->data.r_thread.threadID);
		sprinta0 (str, level, "}\n");
		break;
	case C_INFO_THREADS:
 		sprinta0 (str, level, "info_threads = {\n");
		sprinta0 (str, level, "}\n");
		break;
	case C_SENDTO:
 		sprinta0 (str, level, "sendto = {\n");
		sprinta2 (str, level+1, "command = %d %s\n",
				  STRLEN (tk->data.r_sendto.command),
				  STRING (tk->data.r_sendto.command));
		sprinta0 (str, level, "}\n");
		break;
	case C_TERM_SERVER:
 		sprinta0 (str, level, "terminate_server = {\n");
		sprinta1 (str, level+1, "halt = %d\n",
				  tk->data.r_terminate.halt);
		sprinta0 (str, level, "}\n");
		break;
 	case C_REGISTER_CLIENT:
 		sprinta0 (str, level, "register_client = {\n");
		sprinta1 (str, level+1, "pid = %d\n",
				  tk->data.r_register_client.pid);
		sprinta2 (str, level+1, "filename = %d %s\n",
				  STRLEN (tk->data.r_register_client.filename),
				  STRING (tk->data.r_register_client.filename));
		sprinta0 (str, level, "}\n");
		break;
	default:
		msg_pid_fatal ("%s:%d: Invalid code = %d\n",
					   __FILE__, __LINE__, tk->code);
		break;
	}
	sprinta0 (str, 0, "}\n");
	return mychar_thestr (str);
}




/* -----------------------------------------------------------------
 * Function: char *f0m_tkout_2_jml (const tkout_t *tk)
 *
 * Description: Convert a 'tkout_t' to text
 *
 * Returns:     A dynamically allocated buffer with the text
 */
char *f0m_tkout_2_jml (const tkout_t *tk)
{
	mychar_t *str;
	sprint_data_t spd;

	str = mychar_new ();
	sprinta0 (str, 0, "tkout = {\n");
	sprinta1 (str, 1, "tid = %d\n", tk->tid);
	sprinta2 (str, 1, "str = %d %s\n",
			  STRLEN (tk->str), STRING (tk->str));
	sprinta1 (str, 1, "status = %d\n", tk->status);
	data_2_jml ("rep", tk->rep, str, 1);
	spd.level = 1;
	spd.str = str;
	spd.code = -1;
	sprinta_chain (tk->oob, oob_2_jml, "oob", &spd);
	data_2_jml ("pos", tk->pos, str, 1);
	data_2_jml ("err", tk->err, str, 1);
	sprinta0 (str, 0, "}\n");
	return mychar_thestr (str);
}

