/* *****************************************************************
 *
 *  $Id: f0m_xdr.c,v 1.1 2002/09/27 12:43:02 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/09/16 19:27:31 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */
#include <stdlib.h>
#include "f0m_xdr.h"


f0m_xdr_f f0m_xdr[XDR_LAST];





/* -----------------------------------------------------------------
 * Function:    f0m__getenv_or_default()
 *
 * Description: Get environment variable value, if undef use default 
 *
 * Returns:     []
 */
PRIVATE const char *
f0m__getenv_or_default (const char *env_var, const char *default_value)
{
	const char *val = getenv (env_var);
	if (val == NULL)			/* Environment variable not defined */
		val = default_value;	/* Use default value */
	return val;
}



/* -----------------------------------------------------------------
 * Function:    int f0m_init_xdr ()
 *
 * Description: Initialize the XDR protocol to use (JML || XML)
 *
 * Returns:     The protocol code;
 */
xdr_protocol_t f0m_init_xdr (xdr_protocol_t protocol)
{
	static xdr_protocol_t _protocol = 0;

	if (_protocol == XDR_UNDEF) {
		if (protocol == XDR_UNDEF) {
			const char *env_protocol;
			env_protocol = f0m__getenv_or_default ("FIDDLE_XDR", "JML");
			
			/* Initialize communcation protocol */
			if (strcasecmp (env_protocol, "JML") == 0)
				protocol = XDR_JML;
			else if (strcasecmp (env_protocol, "XML") == 0)
				protocol = XDR_XML;
			else
				msg_pid_fatal ("Invalid XDR protocol \"%s\"\n\n", protocol);
		}		
		
		switch (protocol) {
		case XDR_JML:
			f0m_xdr[TKIN_2_XDR] = (f0m_xdr_f)f0m_tkin_2_jml;
			f0m_xdr[TKOUT_2_XDR] = (f0m_xdr_f)f0m_tkout_2_jml;
			f0m_xdr[XDR_2_TKIN] = (f0m_xdr_f)f0m_jml_2_tkin;
			f0m_xdr[XDR_2_TKOUT] = (f0m_xdr_f)f0m_jml_2_tkout;
			break;
		case XDR_XML:
			f0m_xdr[TKIN_2_XDR] = (f0m_xdr_f)f0m_tkin_2_xml;
			f0m_xdr[TKOUT_2_XDR] = (f0m_xdr_f)f0m_tkout_2_xml;
			f0m_xdr[XDR_2_TKIN] = (f0m_xdr_f)f0m_xml_2_tkin;
			f0m_xdr[XDR_2_TKOUT] = (f0m_xdr_f)f0m_xml_2_tkout;
			break;
		default:
			msg_pid_fatal ("Invalid XDR protocol: %d", protocol);
		}
		_protocol = protocol;
	}
	return _protocol;
}
