/* *****************************************************************
 *
 *  $Id: f0s_dbase.h,v 1.2 2002/08/06 22:03:34 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/08/02 16:34:18 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */
#ifndef __F0_DBASE_H
#define __F0_DBASE_H

#include <stdio.h>
#include "chain.h"



/* ################################################################# */


/* -----------------------------------------------------------------
 * Function:    f0s_dbase_add_cld ()
 *
 * Description: Add a cld + process to the database
 *
 * Returns:     Tid of new process
 */
int f0s_dbase_add_cld (int dbg_pid, int tp_pid,
					   const char *cmd_channel, const char *rep_channel);


/* -----------------------------------------------------------------
 * Function:    f0s_dbase_cleanup ()
 *
 * Description: Remove all entries in the database
 *
 * Returns:     Status
 */
int f0s_dbase_cleanup (void);


/* -----------------------------------------------------------------
 * Function:    int status = f0s_dbase_delete (IN int tid)
 *
 * Description: Remove a debugger from the database
 *
 * Returns:     The status of the operation
 */
int f0s_dbase_delete (int tid);


/* ################################################################# */



/* -----------------------------------------------------------------
 * Function:    int f0s_dbase_list ()
 *
 * Description: List current processes
 *
 * Returns:     Chain with data
 */
chain_t *f0s_dbase_list (void);


/* -----------------------------------------------------------------
 * Function:    int f0s_dbase_is_unused (IN int tid)
 *
 * Description: Check if 'tid' corespond to a unused debugger
 *
 * Returns:     
 */
int f0s_dbase_is_unused (int tid);


/* -----------------------------------------------------------------
 * Function:    int tid = f0s_dbase_find_unused (void)
 *
 * Description: Find an unused debugger (with 'info.tp_pid = -1')
 *
 * Returns:     
 */
int f0s_dbase_find_unused (void);


/* -----------------------------------------------------------------
 * Function:    int tid = f0s_dbase_set_unused (IN int tid)
 *
 * Description: Set 'tid' to unused
 *
 * Returns:     
 */
int f0s_dbase_set_unused (int tid);


/* ################################################################# */


/* -----------------------------------------------------------------
 * Function:    int tid = f0s_dbase_find_by_dbg_pid (IN int dbg_pid)
 *
 * Description: Find the debugger's ID given its PID
 *
 * Returns:     
 */
int f0s_dbase_find_by_dbg_pid (int dbg_pid);


/* -----------------------------------------------------------------
 * Function:    int tid = f0s_dbase_find_by_tp_pid (IN int tp_pid)
 *
 * Description: Find the debugger's ID given the target process's PID
 *
 * Returns:     
 */
int f0s_dbase_find_by_tp_pid (int tp_pid);


/* -----------------------------------------------------------------
 * Function:    int tid = f0s_dbase_find_by_attached (IN int attached)
 *
 * Description: Find the debugger's ID given the target process's PID
 *
 * Returns:     
 */
int f0s_dbase_find_by_attached (int attached);


/* ################################################################# */


/* -----------------------------------------------------------------
 * Function:    int tp_pid = f0s_dbase_get_tp_pid (IN int tid)
 *
 * Description: Get a debugger's target process PID
 *
 * Returns:     The debugger's target process PID
 */
int f0s_dbase_get_tp_pid (int tid);


/* -----------------------------------------------------------------
 * Function:    int status = f0s_dbase_set_tp_pid (IN int tid,
 *												  IN int tp_pid)
 *
 * Description: Set a debugger's target process PID
 *
 * Returns:     The status of the operation
 */
int f0s_dbase_set_tp_pid (int tid, int tp_pid);


/* -----------------------------------------------------------------
 * Function:    int status = f0s_dbase_set_attached (IN int tid,
 *												  IN int attached)
 *
 * Description: Set a debugger's target process PID
 *
 * Returns:     The status of the operation
 */
int f0s_dbase_set_attached (int tid, int attached);


/* ################################################################# */


/* -----------------------------------------------------------------
 * Function:    FILE *rfd = f0s_dbase_get_rfd (IN int tid)
 *
 * Description: Get the channel that allows to receive data from the
 *				    debugger
 *
 * Returns:     The FILE descriptor
 */
FILE *f0s_dbase_get_rfd (int tid);


/* -----------------------------------------------------------------
 * Function:    FILE *wfd = f0s_dbase_get_wfd (IN int tid)
 *
 * Description: Get the channel that allows to send data to the
 *				    debugger
 *
 * Returns:     The FILE descriptor
 */
FILE *f0s_dbase_get_wfd (int tid);


/* -----------------------------------------------------------------
 * Function:    f0s_dbase_get_wfile ()
 *
 * Description: Get the name of the fifo usedto communicate with the CLD
 *
 * Returns:     A pointer to the name
 */
char *f0s_dbase_get_wfile (int tid);


/* -----------------------------------------------------------------
 * Function:    f0s_dbase_get_rfile ()
 *
 * Description: Get the name of the fifo usedto communicate with the CLD
 *
 * Returns:     A pointer to the name
 */
char *f0s_dbase_get_rfile (int tid);


#endif /* __F0_DBASE_H */
