/* *****************************************************************
 *
 *  $Id: f0s_error.c,v 1.2 2002/07/05 15:16:52 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/07/05 15:36:55 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */

#include "f0s_error.h"
#include "config.h"

typedef struct errmsg {
	int        code;
	const char *msg;
} ErrMsg;


static ErrMsg err_msg[] = {
	{E_RTS_SEND_RECEIVE,
	 "Send/Receive failed."},
	{E_RTS_INVALID_TP_PID,
	 "Invalid Target Process (PID)."},
	{E_RTS_NO_CLD_AVAILABLE,
	 "No Component-level Debugger available."},
	{E_RTS_COMMAND_GENERATION,
	 "Command generation failed."},
	{E_RTS_INVALID_TID,
	 "Invalid Target Process Identifier (TID)."},
	{E_RTS_STAT_FILE,
	 "File non-existent or with wrong access permissions."},
	{E_RTS_FILE_PERM,
	 "Wrong file permissions."},
	{E_RTS_ONE_LINE_FUNCTION,
	 "One and only one of 'line' and 'funtion' are allowed."},
	{E_RTS_NO_COMMAND,
	 "No command to send to the CLD."},
	{E_RTS_ASTR_CREATE,
	 "Error creating an Automatic String (memory error)."},
	{E_RTS_INIT_REGEX,
	 "Error compiling the regular expressions to parse GDB output."},
	{E_RTS_INVALID_LINE_NUMBER,
	 "Invalid line number."},
	{E_RTS_INVALID_ARGUMENT,
	 "Invalid argument."},
	{E_RTS_INVALID_OPTION,
	 "Invalid option."},
	{E_RTS_MEMORY_ERROR,
	 "Memory error."},
	{E_RTS_INVALID_SIGNUM,
	 "Invalid signal number."},
	{E_RTS_SIGNAL,
	 "Could not send signal to process."},
	{E_RTS_INVALID_REQID,
	 "Error while processing the request (invalid REQID)."},
	{E_RTS_CANNOT_START_DAEMON,
	 "Couldn't start daemon in remote host (check the '~/.rhosts' file)."},
	{E_RTS_UNKNOWN_OUTPUT,
	 "Unable to parse output."},
	{E_UNKNOWN_CMD_CODE,
	 "Unknown command code."},
	{E_NOT_RUN,
	 "Program is not running."},
	{E_NO_FRAME_SELECTED,
	 "No frame selected."},
	{E_NO_PROCESS,
	 "No such process."},
	{E_NOT_PERMITTED,
	 "Operation not permited."},
	{E_NO_SYMBOL,
	 "No symbols found.  Use \"symbol-file\" to load symbols."},
	{E_BOTTOM,
	 "Bottom of stack reached.  Cannot go down."},
	{E_TOP,
	 "Top of stack reached.  Cannot go up."},
	{E_INVALID_EXPRESSION,
	 "Invalid expression."},
	{E_NO_SYMBOL_TABLE,
	 "No symbol table."},
	{E_NO_DEF_SOURCE_FILE,
	 "No source file."},
	{E_INVALID_BPID,
	 "Invalid breakpoint identifier."},
	{E_INVALID_LINE_NUMBER,
	 "Invalid line number."},
	{E_FUNCTION_NOT_DEFINED,
	 "Function not defined."},
	{E_LINE_OUT_OF_RANGE,
	 "Line number out of range."},
	{E_NO_BOUNDS,
	 "Cannot find bounds of current function."},
	{E_INVALID_SYMBOL,
	 "Invalid symbol in current context."},
	{E_INVALID_DISPLAY_NUMBER,
	 "Invalid display number."},
	{E_NO_DISPLAY_EXPRESSION,
	 "There are no auto-display expressions now."},
	{E_INVALID_ARITHMETIC_OPERATION,
	 "Erroneous arithmetic operation."},
	{E_PARSE_ERROR,
	 "Parse error."},
	{E_TOO_FEW_ARGS,
	 "Too few arguments in function call."},
	{E_NO_STACK,
	 "No stack."},
	{E_NO_BREAK_OR_WATCH,
	 "No breakpoints or watchpoints."},
	{E_NO_LINE_NUM_INFO,
	 "No line number information available."},
	{E_JUNK_AT_END_OF_LINE,
	 "Junk at end of line."},
	{E_INVALID_THREAD_ID,
	 "Invalid thread ID."},
	{E_NO_THREAD_ID,
	 "Please specify a thread ID."},
	{E_NO_THREAD_COMMAND,
	 "Please specify a command following the thread ID."},
	{E_SIGSEGV,
	 "Program received signal SIGSEGV, Segmentation fault."},
	{0,
	 "Invalid command.\nUse \"help\" to list commands.\nUse \"help -s\" to get help on command syntax"}
};


/* -----------------------------------------------------------------
 * Function:    const char *err_msg = f0s_err_msg (code_t code)
 *
 * Description: Return the error text associated with a error code
 *
 * Returns:     Pointer to the error text
 *				'NULL' if not found
 */
const char *f0s_err_msg (code_t code)
{
	int i;
	for (i = 0;  err_msg[i].code != 0;  i++)
		if (code == err_msg[i].code)
			break;
	return err_msg[i].msg;
}

