/* *****************************************************************
 *
 *  $Id: sock_ntop.c,v 1.2 2002/07/07 22:38:34 jml Exp $
 *
 *  Copyright (c) 2000 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/07/07 18:20:08 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description: Taken from "Unix Networking Programming", Vol. I, 
 *				 by W. Richard Stevens
 *
 * -----------------------------------------------------------------
 */

#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>
#include "sock_ntop.h"

/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */


/* #################################################################
 *		PRIVATE functions
 * ################################################################# */


/* #################################################################
 *		PUBLIC functions
 * ################################################################# */


char *sock_ntop(const struct sockaddr *sa, socklen_t salen)
{
    char portstr[7];
    char str[128];		/* Unix domain is largest */

	/* Check arguments */
	assert (sa != NULL);
	assert (salen > 0);
	
	switch (sa->sa_family) {
	case AF_INET: {
		struct sockaddr_in	*sin = (struct sockaddr_in *) sa;

		if (inet_ntop(AF_INET, &sin->sin_addr, str, sizeof(str)) == NULL)
			return(NULL);
		if (ntohs(sin->sin_port) != 0) {
			snprintf(portstr, sizeof(portstr), "/%d", ntohs(sin->sin_port));
			strcat(str, portstr);
		}
		break;
	}
/* end sock_ntop */

#ifdef	IPV6
	case AF_INET6: {
		struct sockaddr_in6	*sin6 = (struct sockaddr_in6 *) sa;

		if (inet_ntop(AF_INET6, &sin6->sin6_addr, str, sizeof(str)) == NULL)
			return(NULL);
		if (ntohs(sin6->sin6_port) != 0) {
			snprintf(portstr, sizeof(portstr), ".%d", ntohs(sin6->sin6_port));
			strcat(str, portstr);
		}
		break;
	}
#endif

#ifdef	AF_UNIX
	case AF_UNIX: {
		struct sockaddr_un	*unp = (struct sockaddr_un *) sa;

			/* OK to have no pathname bound to the socket: happens on
			   every connect() unless client calls bind() first. */
		if (unp->sun_path[0] == 0)
			strcpy(str, "(no pathname bound)");
		else
			snprintf(str, sizeof(str), "%s", unp->sun_path);
		break;
	}
#endif

	default:
		snprintf(str, sizeof(str), "sock_ntop: unknown AF_xxx: %d, len %d",
				 sa->sa_family, salen);
		break;
	}
	return strdup (str);
}

