/* *****************************************************************
 *
 *  $Id: f2m_queues.h,v 1.3 2002/09/27 12:43:03 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/09/26 13:30:48 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description: Prototypes for the enqueue requests and dequeue
 *				 replies services
 *
 *		f2m_queues_init ()		-- Initialization
 *		f2m_queues_terminate ()	-- Termination
 *		f2m_enqueue_request ()	-- Enqueue a request
 *		f2m_dequeue_request ()	-- Dequeue a request
 *		f2m_enqueue_reply ()		-- Enqueue a reply
 *		f2m_dequeue_reply ()		-- Dequeue a reply
 *
 * -----------------------------------------------------------------
 */
#ifndef __F2_QUEUES_H
#define __F2_QUEUES_H


#include <pthread.h>

#include "f0s_tkout.h"
#include "f0m_tkin.h"
#include "glob.h"



/* ID for system services, not associated to a apecific task (yet),
 * such as 'f2m_tids()' and 'fi_attach()' */
#define NO_SERVER -1


typedef unsigned long reqid_t;


typedef struct request_s request_t;
struct request_s {
	const tkin_t  *tk;
	reqid_t        id;
	int            fd;
	pthread_t      thid;
};


typedef struct reply_s reply_t;
struct reply_s {
	const tkout_t *tk;
	reqid_t  id;
};



__BEGIN_DECLS


/* -----------------------------------------------------------------
 * Function:    int f2m_queues_init (void)
 *
 * Description: Initialize queues for requests and replies
 *
 * Returns:      0 if OK
 *				-1 on error
 */
int f2m_queues_init (void);

/* -----------------------------------------------------------------
 * Function:    int f2m_queues_terminate (void)
 *
 * Description: Terminate queues for requests and replies
 *
 * Returns:      0 if OK
 *				-1 on error
 */
int f2m_queues_terminate (int halt);


/* -----------------------------------------------------------------
 * Function:    reqid_t f2m_enqueue_request (const tkin_t *req)
 *
 * Description: Enqueue a request of a service
 *				'req'	 Token with the request code and arguments
 *
 * Returns:     The request ID
 */
reqid_t f2m_enqueue_request (const tkin_t *tk);


/* -----------------------------------------------------------------
 * Function:    request_t *req f2m_dequeue_request (void)
 *
 * Description: Get next entry from request queue
 *
 * Returns:     A structure 'request_t'
 */
request_t *f2m_dequeue_request (void);


/* -----------------------------------------------------------------
 * Function:    int repid = f2m_enqueue_reply (reqid_t id,
 *											   const tkout_t *rep)
 *
 * Description: Enqueue a reply to a service
 *				'id'	The request ID
 *				'rep'	Token with the reply code and arguments
 *
 * Returns:     The status
 */
int f2m_enqueue_reply (reqid_t id, const tkout_t *tk);


/* -----------------------------------------------------------------
 * Function:    tkout_t *tk = f2m_dequeue_reply (reqid_t reqid)
 *
 * Description: Get the reply to the request with ID 'reqid'
 *
 * Returns:     Pointer to the reply ('tkout_t') structure
 */
tkout_t *f2m_dequeue_reply (reqid_t reqid);

__END_DECLS

#endif /* __F2_QUEUES_H */
