/* *****************************************************************
 *
 *  $Id: common.c,v 1.1 2002/08/31 15:23:00 rja Exp $
 *
 *  Copyright (c) 2001 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <01/09/21 17:26:59 rja>
 *
 *  Ricardo Anastacio [jml] <rja@asc.di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */

#include "common.h"

/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */



/* #################################################################
 *		PRIVATE functions
 * ################################################################# */

/*
 * Creates a signature of a Tkout_j constructor
 * (+/-) buffer, a buffer where the string will be placed.
 *       There must be enough space alocated. No buffer
 *       overrun test is perform.
 * (-) a pointer to buffer.
 */
PRIVATE char *IntStrCodeChainDataData_sigV(char *buffer) {
	strcpy(buffer, "(ILjava/lang/String;");
	strcat(buffer, CODE_J_SIG);
	strcat(buffer, CHAIN_J_SIG);
	strcat(buffer, DATA_J_SIG);
	strcat(buffer, DATA_J_SIG);
	strcat(buffer, ")V");

	return buffer;
}
  

/* #################################################################
 *		PUBLIC functions
 * ################################################################# */

PUBLIC void fatalError(const char *msg, int exitCode) {
	fprintf(stderr, msg);
	fflush(stderr);
	exit(exitCode);
}

PUBLIC char *StrCodeChain_sigV(char *buffer) {
	strcpy(buffer, "(");
	strcat(buffer, "Ljava/lang/String;");
	strcat(buffer, CODE_J_SIG);
	strcat(buffer, CHAIN_J_SIG);
	strcat(buffer, ")V");

	return buffer;
}


PUBLIC jobject initTkout_j(JNIEnv *env, indOb_t iObj, repOb_t rObj) {
  jclass    cls;
  jmethodID mid;
  char      buffer[MAX_STR];
	
  cls = (*env)->FindClass(env, TKOUT_J_SIG);
  if(cls == 0) fatalError("file:FindClass:Tkout_j\n", -18);
  mid = (*env)->GetMethodID(env, cls, "<init>", IntStrCodeChainDataData_sigV(buffer));
  if(mid == 0) fatalError("file:GetMethodID:Tkout_j Constructor\n", -2);

#ifdef EBUG
  fprintf(stderr, "Tkout_j vai ser criado com:
  iObj->tid=%d
  iObj->str=0x%x
  iObj->status=0x%x
  iObj->oob=0x%x
  rObj->rep=0x%x
  iObj->pos=0x%x\n", (unsigned int) iObj->tid, (unsigned int) iObj->str, (unsigned int) iObj->status, (unsigned int) iObj->oob, (unsigned int) rObj->rep, (unsigned int) iObj->pos);
  fflush(stderr);
#endif
  
  return (*env)->NewObject(env, cls, mid, iObj->tid, iObj->str, iObj->status, iObj->oob, rObj->rep, iObj->pos);
}




PUBLIC void insertObjInJAVA(JNIEnv *env, jobject javaObj, jobject rcvObj) {
  jclass    cls;
  jmethodID mid;
  jfieldID  fid;
  jobject   data;

  cls = (*env)->GetObjectClass(env, rcvObj);
  if(cls == 0) fatalError("insertObjInJAVA:GetObjectClass:rcvObj\n", -4);
  fid = (*env)->GetFieldID(env, cls, "data", CHAIN_J_SIG);
  if(fid == 0) fatalError("insertObjInJAVA:GetFieldID:rcvObj:data\n", -5);
  data = (*env)->GetObjectField(env, rcvObj, fid);
  if(data == 0) fatalError("insertObjInJAVA:GetObjectField:rcvObj:data\n", -6);

// Lets get the enqueue method of the data (Chain_j) field!
  cls = (*env)->GetObjectClass(env, data);
  if(cls == 0) fatalError("insertObjInJAVA:GetObjectClass:rcvObj:data\n", -4);
  mid = (*env)->GetMethodID(env, cls, "enqueue", "(Ljava/lang/Object;)I");
  if(mid == 0) fatalError("insetObjInJAVA:GetMethodID:enqueue\n", -7);

// Now lets insert the object in the OoB
  (*env)->CallIntMethod(env, data, mid, javaObj);
}

