/* *****************************************************************
 *
 *  $Id: do_OoB.c,v 1.1 2002/08/31 15:23:00 rja Exp $
 *
 *  Copyright (c) 2001 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/08/26 12:42:41 rja>
 *
 *  Ricardo Anastacio [jml] <rja@asc.di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *    Processes the Out of Band structure in the tkout and into a
 * fiddle_j object.
 * -----------------------------------------------------------------
 */

#include "do_OoB.h"

/* #################################################################
 *		#DEFINES, TYPES and GLOBAL (PRIVATE) VARIABLES
 * ################################################################# */

PRIVATE JNIEnv *env;
PRIVATE indOb_t iObj;
PRIVATE tkout_t *tk;
PRIVATE jobject oobData;    /* jobject == Data_j */

/* #################################################################
 *		PRIVATE functions
 * ################################################################# */


#ifdef EBUG
PRIVATE void passo(float f) {
  fprintf(stderr, "do_oob->passo %.2f\n", f);
  fflush(stderr);
}
#endif


/*
 * Auxiliar function for the Chain_walk iterator. It's aplyed to
 * each member of the list Chain_walk traverses, creating for each
 * of this a JAVA object and enqueues it in the OoB member of the
 * current Tkout_j object.
 * (+) obj, the object to convert to JAVA;
 * (+) no_args, extra parameters. These are ignored!
 */
PRIVATE int buildDisplay(void *obj, void *no_args) {
	jobject   java_obj;
	jclass    cls;
	jmethodID mid;
	
// Initialization of the Display fields
	odisplay_t *data     = (odisplay_t *) obj;
	jint      dispid     = data->dispid;
	jstring   expression = (*env)->NewStringUTF(env, data->expression);
	jstring   value      = (*env)->NewStringUTF(env, data->value);


// Creation of the Display_j object
	cls = (*env)->FindClass(env, DISPLAY_J_SIG);
	if(cls==0) fatalError("buildDisplay:FindClass:Display_j\n", -8);
	mid = (*env)->GetMethodID(env, cls, "<init>", "(ILjava/lang/String;Ljava/lang/String;)V");
	if(mid==0) fatalError("buildDisplay:GetMethodID:init\n", -2);

// Construction and instantiation
	java_obj = (*env)->NewObject(env, cls, mid, dispid, expression, value);

// insert the object into the Chain_j structure
	insertObjInJAVA(env, java_obj, oobData);

	return 0;
}



/*
 * Auxiliar function for the Chain_walk iterator. It's aplyed to
 * each member of the list Chain_walk traverses, creating for each
 * of this a JAVA object and enqueues it in the OoB member of the
 * current Tkout_j object.
 * (+) obj, the object to convert to JAVA;
 * (+) no_args, extra parameters. These are ignored!
 */
PRIVATE int buildBphit(void *obj, void *no_args) {
	jobject   java_obj;
	jclass    cls;
	jmethodID mid;
	
// Bphit field initialization
	obphit_t *data    = (obphit_t *) obj;
	jint     bpid     = data->bpid;
	jstring  function = (*env)->NewStringUTF(env, data->function);
	jstring  args     = (*env)->NewStringUTF(env, data->args);
	jstring  file     = (*env)->NewStringUTF(env, data->file);
	jint     line     = data->line;

#ifdef EBUG
	passo(3.61);
#endif
	
// Creation of the Bphit_j object
	cls = (*env)->FindClass(env, BPHIT_J_SIG);
	if(cls==0) fatalError("buildBphit:FindClass:Bphit_j\n", -9);
	mid = (*env)->GetMethodID(env, cls, "<init>", "(ILjava/lang/String;Ljava/lang/String;Ljava/lang/String;I)V");
	if(mid==0) fatalError("buildBphit:GetMethodID:init\n", -2);

#ifdef EBUG
	passo(3.62);
#endif
		
// Construction and instantiation
	java_obj = (*env)->NewObject(env, cls, mid, bpid, function, args, file, line);

#ifdef EBUG
	passo(3.63);
#endif
		
// Insert the object into the Chain_j structure
	insertObjInJAVA(env, java_obj, oobData);

#ifdef EBUG
	passo(3.64);
#endif
	
	return 0;
}






/*
 * Auxiliar function for the Chain_walk iterator. It's aplyed to
 * each member of the list Chain_walk traverses, creating for each
 * of this a JAVA object and enqueues it in the OoB member of the
 * current Tkout_j object.
 * (+) obj, the object to convert to JAVA;
 * (+) no_args, extra parameters. These are ignored!
 */
PRIVATE int buildSignal(void *obj, void *no_args) {
	jobject   java_obj;
	jclass    cls;
	jmethodID mid;
	

// Signal_j field initialization
	osignal_t *data    = (osignal_t *) obj;
	jstring name = (*env)->NewStringUTF(env, data->name);


// Creation of the Signal_j object
	cls = (*env)->FindClass(env, SIGNAL_J_SIG);
	if(cls==0) fatalError("buildSignal:FindClass:Signal_j\n", -10);
	mid = (*env)->GetMethodID(env, cls, "<init>", "(Ljava/lang/String;)V");
	if(mid==0) fatalError("buildSignal:GetMethodID:init\n", -2);

// Construction and instantiation
	java_obj = (*env)->NewObject(env, cls, mid, name);

// Insert the object into the Chain_j structure
	insertObjInJAVA(env, java_obj, oobData);

	return 0;
}




/*
 * Auxiliar function for the Chain_walk iterator. It's aplyed to
 * each member of the list Chain_walk traverses, creating for each
 * of this a JAVA object and enqueues it in the OoB member of the
 * current Tkout_j object.
 * (+) obj, the object to convert to JAVA;
 * (+) no_args, extra parameters. These are ignored!
 */
PRIVATE int buildNewThread(void *obj, void *no_args) {
	jobject   java_obj;
	jclass    cls;
	jmethodID mid;
	

// NewThread_j field initialization
	onewthread_t *data = (onewthread_t *) obj;
	jint         id = data->id;


// Creation of the Signal_j object
	cls = (*env)->FindClass(env, NEWTHREAD_J_SIG);
	if(cls==0) fatalError("buildNewThread:FindClass:NewThread_j\n", -11);
	mid = (*env)->GetMethodID(env, cls, "<init>", "(I)V");
	if(mid==0) fatalError("buildNewThread:GetMethodID:init\n", -2);

// Construction and instantiation
	java_obj = (*env)->NewObject(env, cls, mid, id);

// Insert the object into the Chain_j structure
	insertObjInJAVA(env, java_obj, oobData);

	return 0;
}




/*
 * Auxiliar function for the Chain_walk iterator. It's aplyed to
 * each member of the list Chain_walk traverses, creating for each
 * of this a JAVA object and enqueues it in the OoB member of the
 * current Tkout_j object.
 * (+) obj, the object to convert to JAVA;
 * (+) no_args, extra parameters. These are ignored!
 */
PRIVATE int buildThread(void *obj, void *no_args) {
	jobject   java_obj;
	jclass    cls;
	jmethodID mid;
	

// Thread_j field initialization
	othread_t *data = (othread_t *) obj;
	jint        pid = data->pid;


// Creation of the Signal_j object
	cls = (*env)->FindClass(env, THREAD_J_SIG);
	if(cls==0) fatalError("buildThread:FindClass:Thread_j\n", -12);
	mid = (*env)->GetMethodID(env, cls, "<init>", "(I)V");
	if(mid==0) fatalError("buildThread:GetMethodID:init\n", -2);

// Construction and instantiation
	java_obj = (*env)->NewObject(env, cls, mid, pid);

// Insert the object into the Chain_j structure
	insertObjInJAVA(env, java_obj, oobData);

	return 0;
}




	 
/*
 * Auxiliar function for the Chain_walk iterator. It's aplyed to
 * each structure in the list OoB. For each structure it creates 
 * the corresponding JAVA image and inserts it in the current
 * Tkout_j (the this object).
 * (+) obj, the current structure of the OoB list;
 * (+) no_args, extra parameters. These are ignored!
 */
PRIVATE int process_OoB_structure(void *obj, void *no_args) {
	jclass    cls;
	jmethodID mid;
	jstring   str;
	jobject   code;
	jobject   data;
	data_t    *token = (data_t *) obj;

#ifdef EBUG
	passo(3.1);
#endif
	
	// create the str field of the Data object
	str = (*env)->NewStringUTF(env, token->str);

#ifdef EBUG
	passo(3.2);
#endif
	
// create the code field of the Data object
	cls = (*env)->FindClass(env, CODE_J_SIG);
	if(cls == 0) fatalError("process_OoB_structure: FindClass: Code_j", -1);
	mid = (*env)->GetMethodID(env, cls, "<init>", "(I)V");
	if(mid == 0) fatalError("process_OoB_structure: GetMethodID: Code_j Constructor", -2);
	code = (*env)->NewObject(env, cls, mid, token->code);

#ifdef EBUG
	passo(3.3);
#endif
	
// an empty field for the Data object...
	cls = (*env)->FindClass(env, CHAIN_J_SIG);
	if(cls == 0) fatalError("process_OoB_structure: FindClass: Chain_j", -13);
	mid = (*env)->GetMethodID(env, cls, "<init>", "()V");
	if(mid == 0) fatalError("process_OoB_structure: GetMethodID: Chain_j Constructor", -2);
	data = (*env)->NewObject(env, cls, mid);

	
#ifdef EBUG
	passo(3.4);
#endif
	
// create the Data object to fill the OoB
	cls = (*env)->FindClass(env, DATA_J_SIG);
	if(cls == 0) fatalError("process_OoB_structure: FindClass: Data_j", -14);
	mid = (*env)->GetMethodID(env, cls, "<init>", "(Ljava/lang/String;LCode_j;LChain_j;)V");
	if(mid == 0) fatalError("process_OoB_structure: GetMethodID: Data_j Constructor", -2);
	oobData = (*env)->NewObject(env, cls, mid, str, code, data);

// oobData now exists. We have to fill up the empty field
// with the apropriate type object.	
	switch(token->code) {
	case O_DISPLAY:
#ifdef EBUG
	passo(3.5);
#endif
	
		chain_walk(token->data, buildDisplay, NULL);
		break;
	case O_BPHIT:
#ifdef EBUG
	passo(3.6);
#endif
	
		chain_walk(token->data, buildBphit, NULL);
		break;
	case O_SIGNAL:
#ifdef EBUG
	passo(3.7);
#endif
	
	  chain_walk(token->data, buildSignal, NULL);
	  break;
	case O_NEW_THREAD:
#ifdef EBUG
	passo(3.8);
#endif
	
		chain_walk(token->data, buildNewThread, NULL);
		break;
	case O_SWITCH_TO_THREAD:
#ifdef EBUG
	passo(3.9);
#endif
   
		chain_walk(token->data, buildThread, NULL);
		break;
	default:
		// Some unknown code was found. Stoping traversing, leaving things as they are!
		return -1;
	}

#ifdef EBUG
	passo(3.10);
#endif
	
// Enqueue the new data into the oob
	cls = (*env)->GetObjectClass(env, iObj->oob);
	if(cls == 0) fatalError("process_OoB_structure: GetObjectClass: iObj->oob", -4);
	mid = (*env)->GetMethodID(env, cls, "enqueue", "(Ljava/lang/Object;)I");
	if(mid == 0) fatalError("process_OoB_structure: GetMethodID: enqueue", -7);
	(*env)->CallIntMethod(env, iObj->oob, mid, oobData);

#ifdef EBUG
	passo(3.11);
#endif
	
	return 0;
}


/* #################################################################
 *		PUBLIC functions
 * ################################################################# */

PUBLIC void process_OoB(JNIEnv *envArg, indOb_t iObjArg, tkout_t *tkArg) {
	jclass    cls;
	jmethodID mid;
	
#ifdef EBUG
	passo(1);
#endif
	
// set the global variables
	env = envArg;
	iObj = iObjArg;
	tk = tkArg;

#ifdef EBUG
	passo(2);
#endif
	
// create the OutOfBand Fiddle_j object
	cls = (*env)->FindClass(env, CHAIN_J_SIG);
	if(cls == 0) fatalError("process_OoB: FindClass: Chain_j", -13);
	mid = (*env)->GetMethodID(env, cls, "<init>", "()V");
	if(mid == 0) fatalError("process_OoB: GetMethodID: Chain_j Constructor", -2);
	iObj->oob = (*env)->NewObject(env, cls, mid);

#ifdef EBUG
	passo(3);
#endif
	
// traverse the C OutOfBand out of the tkout_t structure
// processing each token by its type	
	if(chain_walk(tk->oob, process_OoB_structure, NULL) != NULL)
	  fatalError("Error processing Out of Band structure. Traversing
 not completed. Program canceled.", -15);

#ifdef EBUG
	passo(4);
#endif
}

