/* *****************************************************************
 *
 *  $Id: cmds.c,v 1.4 2002/09/27 12:43:03 jml Exp $
 *
 *  Copyright (c) 1999 by Dep.Informatica of the FCT/UNL
 *  Time-stamp: <02/09/26 11:32:41 jml>
 *
 *  Joao Lourenco <Joao.Lourenco@di.fct.unl.pt>
 *
 *  Parallel and Distributed Processing Group
 *  Departamento de Informatica
 *  Faculdade de Ciencias e Tecnologia
 *  Universidade Nova de Lisboa
 *
 * -----------------------------------------------------------------
 *
 *  Description:
 *
 * -----------------------------------------------------------------
 */

#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

#include "fx_lib.h"

#include "test_output.h"
#include "cmds.h"
#include "error.h"
#include "config.h"



#define MAX_DEFAULT_TID 10
#define MAX_BUF 512

char default_tid[MAX_DEFAULT_TID];



#define CHECK_N_ARGS(nargs)												\
    if (argc != nargs) return f0s_tkout_new (CE_INVALID_NARGS);			\
	if (argv[0] != NULL) {												\
		tid = my_atoi (argv[0]);										\
		if (tid == INT_MIN) return f0s_tkout_new (CE_INVALID_ARGUMENT);		\
	}																	\
	else																\
		tid = -1;




PUBLIC const DbgCommand cmd_table[] = {
    {cmd_help,
	 ", \t",
     "help",
     "Give help on command. Use \"help -s\" for help on command syntax.",
     "       %s <-s>? <command>?"
    },
    {cmd_settid,
	 ", \t",
     "settid",
     "Set default tid",
     "<tid>  %s"
    },
    {cmd_quit,
	 ", \t",
     "quit",
     "Quit the debugging session.",
     "       %s"
    },
    {cmd_halt,
	 ", \t",
     "halt",
     "Terminate the debugging session.",
     "       %s"
    },
#ifndef f0s
    {cmd_signal,
	 ", \t",
     "signal",
     "Send a signal to a process.",
     "<tid>  %s <signum>"
    },
#endif
    {cmd_tids,
	 ", \t",
     "tids",
     "List the Target Process Identifiers (TIDs) available.",
     "       %s <-a>?"
    },
    {cmd_add_debugger,
	 ", \t",
     "add-debugger",
     "Incorporate an already running debugger (node debugger) into fiddle.",
#if f0s || f0m
     "<tid>? %s <cld_pid> <tp_pid> <cmd_channel> <rep_channel>"
#else
     "<tid>? %s <host_name:>?<cld_pid> <tp_pid> <cmd_channel> <rep_channel>"
#endif
    },
    {cmd_kill,
	 ", \t",
     "kill",
     "Kill a running process.",
     "       %s"
    },
    {cmd_attach,
	 ", \t",
     "attach",
     "Attach to a running process.",
#if f0s || f0m
     "       %s <pid>"
#else
	 "       %s <host_name:>?<pid>"
#endif
    },
    {cmd_detach,
	 ", \t",
     "detach",
     "Detach from a process.  The process will be left running.",
     "<tid>  %s"
    },
    {cmd_symbol_file,
	 ", \t",
     "symbol-file",
     "Load the program symbols from 'file'.",
#if f0s || f0m
     "<tid>  %s <exec-file-name>"
#else
	  "<tid>  %s <<host_name:>?exec-file-name>"
#endif
    },
    {cmd_file,
	 ", \t",
     "file",
     "Load the image of an executable to memory.  Use \"run\" to\n\
                 execute the program.",
#if f0s || f0m
     "<tid>? %s <exec-file-name>"
#else
	 "<tid>? %s <host_name:>?<exec-file-name>"
#endif
    },
    {cmd_run,
	 ", \t",
     "run",
     "Start running an already loaded program.",
     "<tid>  %s <<arg> ...>?."
    },
    {cmd_step,
	 ", \t",
     "step",
     "Execute until next instruction stepping into function calls.",
     "<tid>  %s"
    },
    {cmd_next,
	 ", \t",
     "next",
     "Execute until next instruction stepping over function calls.",
     "<tid>  %s"
    },
    {cmd_continue,
	 ", \t",
     "continue",
     "Continue the execution of a stopped process.",
     "<tid>  %s"
    },
    {cmd_finish,
	 ", \t",
     "finish",
     "Continue the execution of a process until leaving the current\n\
                 stack frame.",
     "<tid>  %s"
    },
    {cmd_call,
	 " \t",
     "call",
     "Execute a function.",
     "<tid>  %s <fucntion_name> (<arg1, ...>?)"
    },
    {cmd_break,
	 ", \t",
     "break",
     "Set a break point in the function/line number specified.",
     "<tid>  %s <<sourcefile:>?<line_number|function_name>>"
    },
    {cmd_delete,
	 ", \t",
     "delete",
     "Delete a break point.",
     "<tid>  %s <breakpoint_number>"
    },
    {cmd_tty,
	 ", \t",
     "tty",
     "Set a terminal for future runs of program being debugged.",
     "<tid>  %s <character-device>"
    },
    {cmd_set_variable,
	 ", \t",
     "set-variable",
     "Sets a variable to the result of the evaluation of an expression.",
     "<tid>  %s <variable> = <expression>"
    },
    {cmd_evaluate,
	 ", \t",
     "evaluate",
     "Evaluate an expression in the current context.",
     "<tid>  %s <expression>"
    },
    {cmd_display,
	 ", \t",
     "display",
     "Display an expression's value every time the process stops.\n\
                 Without 'expression' display all current expression\n\
                 values.",
     "<tid>  %s <expression>?"
    },
    {cmd_undisplay,
	 ", \t",
     "undisplay",
     "Stop displaying an expression.",
     "<tid>  %s <display_id>"
    },
    {cmd_list,
	 ", \t",
     "list",
     "List a source file.",
     "<tid>  %s <<sourcefile:>?<sart_line|function_name>> <n_lines>?"
    },
    {cmd_info_line,
	 ", \t",
     "info-line",
     "Get information about the line specified.",
     "<tid>  %s <<sourcefile:>?<line_number|function_name>>"
    },
    {cmd_info_program,
	 ", \t",
     "info-program",
     "Get information about a running program.",
     "<tid>  %s"
    },
    {cmd_info_stack,
	 ", \t",
     "info-stack",
     "Get information about the stack frames.",
     "<tid>  %s"
    },
    {cmd_info_locals,
	 ", \t",
     "info-locals",
     "Get name and value of variables in current stack frame.",
     "<tid>  %s"
    },
    {cmd_info_args,
	 ", \t",
     "info-args",
     "Get name and value of function arguments.",
     "<tid>  %s"
    },
    {cmd_up,
	 ", \t",
     "up",
     "Go up in the stack frame list.",
     "<tid>  %s"
    },
    {cmd_down,
	 ", \t",
     "down",
     "Go down in the stack frame list.",
     "<tid>  %s"
    },
    {cmd_frame,
	 ", \t",
     "frame",
     "Select a specific frame from frame list.",
     "<tid>  %s <frame>"
    },
    {cmd_info_break,
	 ", \t",
     "info-break",
     "Get information about the current breakpoints.",
     "<tid>  %s"
    },
    {cmd_info_display,
	 ", \t",
     "info-display",
     "Get information about the current display expressions.",
     "<tid>  %s"
    },
    {cmd_info_threads,
	 ", \t",
     "info-threads",
     "Get information about the currently available threads.",
     "<tid>  %s"
    },
    {cmd_thread,
	 ", \t",
     "thread",
     "Set the default target thread.",
     "<tid>  %s <th_id>"
    },
    {cmd_sendto,
	 ", \t",
     "sendto",
     "Send command directly to the debugger with no parsing.",
     "<tid>  %s <command>"
    },
#if f2m || f3m
	{cmd_clients,
	 ", \t",
     "clients",
     "List current Fiddle clients.",
     "<tid>  %s <-f filename>? <-m host>?"
    },
#endif
    {0,  0,  0,  0},
};



/* -----------------------------------------------------------------
 * Function:    int number = my_atoi (const char *str)
 *
 * Description: Convert a string into an integer.
 *
 * Returns:     The number or INT_MIN on error
 */
PUBLIC int
my_atoi (const char *str)
{
    char *endptr;
    int number;
	if (str == NULL)
		return INT_MIN;
	number = (int) strtol (str, &endptr, 0);
	assert (endptr != NULL);
    return (*endptr == '\0') ? number : INT_MIN;
}





/* #################################################################
   argv[0] = "<tid>" || NULL
   argv[1] = "<command>"
   argv[i] = "<arg_i>"     for all (1 < i < argc)
   ################################################################# */



/* ================================================================= 
 * Process the 'help <-s>? <command>?" command
 */
PUBLIC tkout_t *
cmd_help (int argc, const char *argv[])
{
    /* If no arguments, present description of commands */
    /* If option "-s" is given present the syntax of commands */
    /* With argument, restrict description/syntax to that command */
    int i;

	if (argv[0] != NULL)
		printf ("WARNING: <tid> ignored\n");
    switch (argc) {
    case 2: /* No argument */
        puts("\
-----------------------------------------------------------------\n\
Use 'help [-s] command' to get detailed help\n\
-----------------------------------------------------------------");
        printf ("%-12s %s\n", "COMMAND", "ARGUMENTS");
        for (i = 0;  cmd_table[i].cmd_funct != 0;  i++)
            printf ("%-12s %s\n",
                    cmd_table[i].cmd_str, cmd_table[i].cmd_help1);
        puts ("\n\
-----------------------------------------------------------------");
        break;
    case 3: /* Help on syntax or on command */
        if (strcmp (argv[2], "-s") == 0) {
            /* Help on syntax */
            puts("\
-----------------------------------------------------------------");
            printf ("%-14s\n", "COMMAND SYNTAX");
            for (i = 0;  cmd_table[i].cmd_funct != 0;  i++) {
                printf (cmd_table[i].cmd_help2, cmd_table[i].cmd_str);
				puts ("");
			}
            puts ("\n\
-----------------------------------------------------------------");
        }
        else {
            /* Help on command */
            for (i = 0;  cmd_table[i].cmd_funct != 0;  i++)
                if (strcmp (argv[2], cmd_table[i].cmd_str) == 0)
                    break;
            if (cmd_table[i].cmd_funct == 0)
                return f0s_tkout_new (CE_NO_HELP);
            else {
            puts ("\n\
-----------------------------------------------------------------");
                printf ("%-12s %s\n",
                        cmd_table[i].cmd_str, cmd_table[i].cmd_help1);
            puts ("\n\
-----------------------------------------------------------------");
            }
        }
        break;
    case 4: /* Help on or on command syntax */
        if (strcmp (argv[2], "-s") != 0)
            return f0s_tkout_new (CE_INVALID_ARGUMENT);
        for (i = 0;  cmd_table[i].cmd_funct != 0;  i++)
            if (strcmp (argv[3], cmd_table[i].cmd_str) == 0)
                break;
        if (cmd_table[i].cmd_funct == 0)
            return f0s_tkout_new (CE_NO_HELP);
        else {
            printf ("\nSYNTAX: ");
			printf (cmd_table[i].cmd_help2, cmd_table[i].cmd_str);
			puts ("");
		}
        break;
    default:
        /* Impossible to be reached (I hope) */
        return f0s_tkout_new (CE_NO_HELP);
    }
    return f0s_tkout_new (CE_OK);	/* No error */
}


/* ================================================================= 
 * Processes the "quit" command. 
 * Leave gracefully the debuggin console
 * This function does not return;
 */
PUBLIC tkout_t *
cmd_quit (int argc, const char *argv[])
{
	int tid;
	
	CHECK_N_ARGS (2);
#ifdef f2m
    fx_terminate(0);
#else
    fx_terminate();
#endif
    printf ("\nGoodbye and thanks for using this program\n");
    exit (0);
}



/* ================================================================= 
 * Processes the "halt" command. 
 * Leave gracefully the debuggin console
 * This function does not return;
 */
PUBLIC tkout_t *
cmd_halt (int argc, const char *argv[])
{
	int tid;
	
	CHECK_N_ARGS (2);
#ifdef f2m
    fx_terminate(1);
#else
    fx_terminate();
#endif
    printf ("\nGoodbye and thanks for using this program\n");
    exit (0);
}




/* ================================================================= 
 * Process the "<tid> settid" command
 */
PUBLIC tkout_t *
cmd_settid (int argc, const char *argv[])
{
	int tid;
	/* Check arguments */
	CHECK_N_ARGS (2);
	if (tid <= 0)
		argv[0] = "";
	strncpy (default_tid, argv[0], MAX_DEFAULT_TID);
	default_tid[MAX_DEFAULT_TID - 1] = '\0';
    return f0s_tkout_new (CE_OK);	/* No error */
}

/* ================================================================= 
 * Processes the "signal <tid> <signum>" command. 
 * Send a signal to a process
 */
PUBLIC tkout_t *
cmd_signal (int argc, const char *argv[])
{
	int tid;
    int signum;

    /* Check arguments */
	CHECK_N_ARGS (3);
    signum = my_atoi (argv[2]);
    if (signum < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Eexecute the command */
    return fx_signal (tid, signum);
}


/* ================================================================= 
 * Processes the "tids <-a>?" command.
 * List all the available Target Processes Identifiers
 */
PUBLIC tkout_t *
cmd_tids (int argc, const char *argv[])
{
	int i;
	char args[MAX_BUF];
	tkout_t *tk;

    /* Check arguments */
    if (argc < 2) return f0s_tkout_new (CE_INVALID_NARGS);
	if (argv[0] != NULL) f0s_tkout_new (CE_INVALID_ARGUMENT);

	/* Join the 2nd and following arguments into just one string */
	strcpy (args, "");
	for (i = 2;  i < argc;  i++) {
		strcat (args, argv[i]);
		strcat (args, " ");
	}
    
    /* Eexecute the command */
    tk = fx_tids (args);
	return tk;
}


#if f2m | f3m
/* ================================================================= 
 * Processes the "clients  <-f filename>? <-m host>?" command.
 * List current Fiddle clients.
 */
PUBLIC tkout_t *
cmd_clients (int argc, const char *argv[])
{
	int i;
	char args[MAX_BUF];
	tkout_t *tk;

    /* Check arguments */
    if (argc < 2) return f0s_tkout_new (CE_INVALID_NARGS);
	if (argv[0] != NULL) f0s_tkout_new (CE_INVALID_ARGUMENT);

	/* Join the 2nd and following arguments into just one string */
	strcpy (args, "");
	for (i = 2;  i < argc;  i++) {
		strcat (args, argv[i]);
		if (i < argc - 1)
			strcat (args, " ");
	}
    
    /* Eexecute the command */
    tk = fx_clients (args);
	return tk;
}
#endif


/* ================================================================= 
 * Processes the "add-debugger" command. 
 * Incorporate an aready running debugger (node debugger) into fiddle
 * ### This command doesn't use the defult tid ###
 */
PUBLIC tkout_t *
cmd_add_debugger (int argc, const char *argv[])
{
	char *save;
	int tid;
	const char *host;
	const char *cld_pid_str;
	int cld_pid;
	int tp_pid;
	const char *cmd_channel;
	const char *rep_channel;

    /* Check arguments */
	if (argc != 6) return f0s_tkout_new (CE_INVALID_NARGS);
	if (argv[0] != NULL) {
		tid = my_atoi (argv[0]);
		if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);
	}
	else
		tid = -1;

    /* Extract the 'host' and the 'cld_pid' */
    host = strtok_r ((char *)argv[2], ":", &save);
    cld_pid_str = strtok_r (NULL, ":", &save);
#if f0s || f0m
    if (cld_pid_str != NULL)
		return f0s_tkout_new (CE_INVALID_ARGUMENT);
#endif
    if (cld_pid_str == NULL) {
        /* No host specified */
        cld_pid_str = host;  
        host = "localhost";
    }
	
    /* Extract the 'pid' of the node debugger */
    cld_pid = my_atoi (cld_pid_str);
	if (cld_pid <= 0) return f0s_tkout_new (E_RTS_INVALID_ND_PID);;

    /* Extract the 'pid' of the target process */
	tp_pid = my_atoi (argv[3]);
	if (tp_pid < 0) return f0s_tkout_new (E_RTS_INVALID_TP_PID);;

	/* Get command and reply channels */
	cmd_channel = argv[4];
	assert (cmd_channel != NULL);
	rep_channel = argv[5];
	assert (rep_channel != NULL);
	
    /* Eexecute the command */
#if f0s || f0m
    return fx_add_debugger (cld_pid, tp_pid, cmd_channel, rep_channel);
#else
	return fx_add_debugger (host, cld_pid, tp_pid, cmd_channel, rep_channel);
#endif
}


/* ================================================================= 
 * Processes the "kill <tid>" command. 
 * Kill a running process
 */
PUBLIC tkout_t *
cmd_kill (int argc, const char *argv[])
{
	int tid;
	
    /* Check arguments */
	CHECK_N_ARGS (2);

    /* Eexecute the command */
    return fx_kill (tid);
}


/* ================================================================= 
 * (f0s, f0m) Processes the "attach <tp_pid>" command. 
 * (f1m, f2m) Processes the "attach <host_name>:<tp_pid>" command. 
 * Attach to a running process
 */
PUBLIC tkout_t *
cmd_attach (int argc, const char *argv[])
{
	char *save;
    int tp_pid;
	const char *host;
	const char *tp_pid_str;
    int tid;
    char p[MAX_BUF];
    tkout_t *tk;

    /* Check arguments */
    if (argc != 3) return f0s_tkout_new (CE_INVALID_NARGS);
	if (argv[0] != NULL) f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Extract the 'host' and the 'tp_pid' */
    host = strtok_r ((char *)argv[2], ":", &save);
    tp_pid_str = strtok_r (NULL, ":", &save);
#if f0s || f0m
    if (tp_pid_str != NULL)
		return f0s_tkout_new (CE_INVALID_ARGUMENT);
#endif
    if (tp_pid_str == NULL) {
        /* No host specified */
        tp_pid_str = host;  
        host = "localhost";
    }
	
    /* Extract the 'sourcefile' and the 'pid' of the target process */
    tp_pid = my_atoi (tp_pid_str);
	if (tp_pid <= 0) return f0s_tkout_new (E_RTS_INVALID_TP_PID);;

    /* Eexecute the command */
#if f0s || f0m
    tk = fx_attach (tp_pid);
#else
	tk = fx_attach (host, tp_pid);
#endif
	
    /* Look for errors */
    if (tk->status != 0  ||  tk->err != NULL || tk->pos != NULL)
        /* If there was an error or the symvols were already loaded
		 * return immediately */
        return tk;

    /* Probably the command succeeded, so let's print the relevant 
     * data and then  try to load symbols*/
    test_reply (tk->rep);
    test_oob (tk->oob);

    /* Delete the temporary output token */
    tid = ((rtids_t *)(chain_first_data (tk->rep->data)))->gtid;
    f0s_tkout_delete (tk);

    /* Try to load symbols */
#if __linux__
#define FORMAT_PROC "/proc/%d/exe"
#else
#error PLEASE specify how to find the image in the '/proc' filesystem
#endif /* __linux__ */
    sprintf (p, FORMAT_PROC,  tp_pid);
    printf ("    Trying to load symols from '%s'...\n", p);
#if f0s || f0m
    tk = fx_symbol_file (tid, p);
#else
    tk = fx_symbol_file (tid, host, p);
#endif

    return tk;
}


/* ================================================================= 
 * Processes the "detach <tid>" command. 
 * Detach from a process
 */
PUBLIC tkout_t *
cmd_detach (int argc, const char *argv[])
{
	int tid;
	
    /* Check arguments */
	if (argc != 2) return f0s_tkout_new (CE_INVALID_NARGS);
	tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);
	

    /* Eexecute the command */
    return fx_detach (tid);
}


/* ================================================================= 
 * Processes the "symbol-file <tid> <exec-file-name>" command. 
 * Load the program symbols
 */
PUBLIC tkout_t *
cmd_symbol_file (int argc, const char *argv[])
{
	int tid;
	char *save;
	const char *host;
	const char *filename;

    /* Check arguments */
	if (argc != 3) return f0s_tkout_new (CE_INVALID_NARGS);
	tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

	/* Extract the 'host' and the 'filename' */
	host = strtok_r ((char *)argv[2], ":", &save);
	filename = strtok_r (NULL, ":", &save);
#if f0s || f0m
    if (filename != NULL)
		return f0s_tkout_new (CE_INVALID_ARGUMENT);
#endif
	if (filename == NULL) {
		/* No file specified */
		filename = host;  
		host = "localhost";
	}
    
    /* Eexecute the command */
#if f0s || f0m
    return fx_symbol_file (tid, filename);
#else
	return fx_symbol_file (tid, host, filename);
#endif
}


/* ================================================================= 
 * Processes the "file <tid> <exec-file-name>" command. 
 * Load the program into memory
 * ### This command doesn't use the defult tid ###
 */
PUBLIC tkout_t *
cmd_file (int argc, const char *argv[])
{
	char *save;
	int tid;
	const char *host;
	const char *filename;

    /* Check arguments */
	if (argc != 3) return f0s_tkout_new (CE_INVALID_NARGS);
	if (argv[0] != NULL) {
		tid = my_atoi (argv[0]);
		if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);
	}
	else
		tid = -1;
	
	/* Extract the 'host' and the 'filename' */
	host = strtok_r ((char *)argv[argc - 1], ":", &save);
	filename = strtok_r (NULL, ":", &save);
#if f0s || f0m
	if (filename != NULL)
		return f0s_tkout_new (CE_INVALID_ARGUMENT);
#endif
	if (filename == NULL) {
		/* No file specified */
		filename = host;  
		host = "localhost";
	}
	
    /* Eexecute the command */
#if f0s || f0m
    return fx_file (tid, filename);
#else
	return fx_file (tid, host, filename);
#endif
}


/* ================================================================= 
 * Processes the "run <tid>" command. 
 * Starts the execution of a process
 */
PUBLIC tkout_t *
cmd_run (int argc, const char *argv[])
{
    int tid;
	char args[MAX_BUF];
	int i;

    /* Check arguments */
	if (argc < 2) return f0s_tkout_new (CE_INVALID_NARGS);
	tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);
	
    /* Buil the command string in a 'string' */
    strcpy (args, "");
    for (i = 2;  i < argc; i++) {
        strcat (args,  " ");
        strcat (args, argv[i]);
    }

    /* Eexecute the command */
    return fx_run (tid, args);
}


/* ================================================================= 
 * Processes the "step <tid>" command. 
 * Step the execution (into function called)
 */
PUBLIC tkout_t *
cmd_step (int argc, const char *argv[])
{
    int tid;

    /* Check arguments */
	if (argc < 2) return f0s_tkout_new (CE_INVALID_NARGS);
	tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Eexecute the command */
    return fx_step (tid);
}


/* ================================================================= 
 * Processes the "next <tid>" command. 
 * Step the execution (over function calls)
 */
PUBLIC tkout_t *
cmd_next (int argc, const char *argv[])
{
    int tid;

    /* Check arguments */
	if (argc < 2) return f0s_tkout_new (CE_INVALID_NARGS);
	tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Eexecute the command */
    return fx_next (tid);
}


/* ================================================================= 
 * Processes the "continue <tid>" command. 
 * Continue the execution of a process
 */
PUBLIC tkout_t *
cmd_continue (int argc, const char *argv[])
{
    int tid;

    /* Check arguments */
	if (argc < 2) return f0s_tkout_new (CE_INVALID_NARGS);
	tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Eexecute the command */
    return fx_continue (tid);
}


/* ================================================================= 
 * Processes the "finish <tid>" command. 
 * Continue the execution of a process until leaving the current stack 
 *     frame
 */
PUBLIC tkout_t *
cmd_finish (int argc, const char *argv[])
{
    int tid;

    /* Check arguments */
	if (argc < 2) return f0s_tkout_new (CE_INVALID_NARGS);
	tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Eexecute the command */
    return fx_finish (tid);
}


/* ================================================================= 
 * Processes the "call <tid> <fucntion_name> (<arg1, ...>?)" command. 
 * Call a function and print its result
 */
PUBLIC tkout_t *
cmd_call (int argc, const char *argv[])
{
    int tid;
	int i;
	char function[MAX_BUF];
	tkout_t *tk;

    /* Check arguments */
	if (argc < 3) return f0s_tkout_new (CE_INVALID_NARGS);
	tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Buil the command string in an 'string' */
    strcpy (function, argv[2]);
    for (i = 3;  i < argc; i++) {
        strcat (function,  " ");
        strcat (function, argv[i]);
    }

    /* Eexecute the command */
    tk = fx_call (tid, function);
    return tk;
}


/* ================================================================= 
 * Processes the "break <tid> <<sourcefile:>?<line_number|function_name>>"
 * command.
 */
PUBLIC tkout_t *
cmd_break (int argc, const char *argv[])
{
    char *save;
    const char *lf;
    int tid;
    const char *file;
    int line;
    const char *function;

    /* Check arguments */
	if (argc != 3) return f0s_tkout_new (CE_INVALID_NARGS);
	tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);
    
    /* Extract the 'sourcefile', the 'line_number' and the 'function_name' */
    file = strtok_r ((char *)argv[2], ":", &save);
    lf = strtok_r (NULL, ":", &save);
    if (lf == NULL) {
        /* No file specified */
        lf = file;      
        file = NULL;
    }
    line = my_atoi (lf);
    if (line == INT_MIN) {
        /* Function name specified */
        line = -1;                              
        function = lf;
    }
    else
        /* Line number specified */
        function = NULL;                

    /* Eexecute the command */
    return fx_break (tid, file, line, function);
}


/* ================================================================= 
 * Processes the "delete <tid> <bpid>" command. 
 * Delete a breakpoint
 */
PUBLIC tkout_t *
cmd_delete (int argc, const char *argv[])
{
    int tid;
    int bpid;

    /* Check arguments */
	if (argc != 3) return f0s_tkout_new (CE_INVALID_NARGS);
	tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);
    bpid = my_atoi (argv[2]);
    if (bpid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Eexecute the command */
    return fx_delete (tid, bpid);
}


/* ================================================================= 
 * Processes the "set-variable <tid> <var_name> = <expression>" command. 
 * Evaluate an expression and set the variable contents
 */
PUBLIC tkout_t *
cmd_set_variable (int argc, const char *argv[])
{
    int tid;
	int i;
	char buffer[MAX_BUF];
	char *variable;
	char *expression;
	tkout_t *tk;

    /* Check arguments */
    if (argc < 3) return f0s_tkout_new (CE_INVALID_NARGS);
    tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

	/* Join the 3rd and following arguments into just one string */
	strcpy (buffer, "");
	for (i = 2;  i < argc;  i++) {
		strcat (buffer, argv[i]);
		strcat (buffer, " ");
	}
	
	/* Now, we must break this string in two,
	 * using the first '=' as separator */
	variable = buffer;
	expression = strchr (buffer, '=');
	if (expression == NULL) return f0s_tkout_new (CE_INVALID_NARGS);
	*expression = '\0';
	++expression;
	
	/* Eexecute the command */
	tk = fx_set_variable (tid, variable, expression);
	return tk;
}


/* ================================================================= 
 * Processes the "evaluate <tid> <expression>" command. 
 * Evaluate an expression 
 */
PUBLIC tkout_t *
cmd_evaluate (int argc, const char *argv[])
{
    int tid;
	int i;
	char args[MAX_BUF];
	tkout_t *tk;

    /* Check arguments */
    if (argc < 3) return f0s_tkout_new (CE_INVALID_NARGS);
    tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

	/* Join the 3rd and following arguments into just one string */
	strcpy (args, "");
	for (i = 2;  i < argc;  i++) {
		strcat (args, argv[i]);
		strcat (args, " ");
	}

    /* Eexecute the command */
    tk = fx_evaluate (tid, args);
	return tk;
}


/* ================================================================= 
 * Processes the "info-locals <tid>" command.
 */
PUBLIC tkout_t *
cmd_info_locals (int argc, const char *argv[])
{
    int tid;

    /* Check arguments */
    if (argc != 2) return f0s_tkout_new (CE_INVALID_NARGS);
    tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Eexecute the command */
    return fx_info_locals (tid);
}


/* ================================================================= 
 * Processes the "info-args <tid>" command.
 */
PUBLIC tkout_t *
cmd_info_args (int argc, const char *argv[])
{
    int tid;

    /* Check arguments */
    if (argc != 2) return f0s_tkout_new (CE_INVALID_NARGS);
    tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Eexecute the command */
    return fx_info_args (tid);
}


/* ================================================================= 
 * Processes the "display <tid> <expression>?" command. 
 * Display an expression every time the process stops
 */
PUBLIC tkout_t *
cmd_display (int argc, const char *argv[])
{
    int tid;
	int i;
	char args[MAX_BUF];
	tkout_t *tk;

    /* Check arguments */
    if (argc < 2) return f0s_tkout_new (CE_INVALID_NARGS);
    tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

	/* Join the 3rd and following arguments into just one string */
	strcpy (args, "");
	for (i = 2;  i < argc;  i++) {
		strcat (args, argv[i]);
		strcat (args, " ");
	}

    /* Eexecute the command */
    tk = fx_display (tid, args);
	return tk;
}


/* ================================================================= 
 * Processes the "undisplay <tid> <dispid>" command. 
 * Undisplay a breakpoint
 */
PUBLIC tkout_t *
cmd_undisplay (int argc, const char *argv[])
{
    int tid;
    int dispid;

    /* Check arguments */
    if (argc != 3) return f0s_tkout_new (CE_INVALID_NARGS);
    tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);
    dispid = my_atoi (argv[2]);
    if (dispid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Eexecute the command */
    return fx_undisplay (tid, dispid);
}


/* ================================================================= 
 * Processes the "<tid> tty <exec-tty-name>" command. 
 * Load the program into memory
 */
PUBLIC tkout_t *
cmd_tty (int argc, const char *argv[])
{
    int tid;

    /* Check arguments */
    if (argc != 3) return f0s_tkout_new (CE_INVALID_NARGS);
    tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Eexecute the command */
    return fx_tty (tid, argv[2]);
}


/* ================================================================= 
 * Processes the
 *     "list <tid> <<sourcefile:>?<sart_line|function_name>> <n_lines>?"
 *     command.
 */
PUBLIC tkout_t *
cmd_list (int argc, const char *argv[])
{
    char *save;
    const char *lf;
    int tid;
    const char *file;
    int line;
    const char *function;
    int nlines;

    /* Check arguments */
    if (argc < 3 ||  argc > 4) return f0s_tkout_new (CE_INVALID_NARGS);
    tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Extract the 'sourcefile', the 'line_number' and the 'function_name' */
    file = strtok_r ((char *)argv[2], ":", &save);
    lf = strtok_r (NULL, ":", &save);
    if (lf == NULL) {
        /* No file specified */
        lf = file;  
        file = NULL;
    }
    line = my_atoi (lf);
    if (line < 0) {
        /* Function name specified */
        line = -1;              
        function = lf;
    }
    else
        /* Line number specified */
        function = NULL;

    /* Extract the number of lines to list */
    if (argc == 4) {
        nlines = my_atoi (argv[3]);
        if (nlines < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);
    }
    else
        nlines = -1;

    /* Eexecute the command */
    return fx_list (tid, file, line, function, nlines);
}


/* ================================================================= 
 * Processes the "info-line <tid> <<sourcefile:>?<line_number|function_name>>"
 * command.
 */
PUBLIC tkout_t *
cmd_info_line (int argc, const char *argv[])
{
    char *save;
    const char *lf;
    int tid;
    const char *file;
    int line;
    const char *function;

    /* Check arguments */
    if (argc < 2 || argc > 3) return f0s_tkout_new (CE_INVALID_NARGS);
    tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);
    
	if (argc == 2) {
		/* No arguments */
		file = function = NULL;
		line = -1;
	}
	else {
		/* Extract the 'sourcefile', the 'line_number' and the
		 * 'function_name' */
		file = strtok_r ((char *)argv[2], ":", &save);
		lf = strtok_r (NULL, ":", &save);
		if (lf == NULL) {
			/* No file specified */
			lf = file;      
			file = NULL;
		}
		if (lf == NULL) {
			line = -1;                              
			function = NULL;
		}
		else {
			line = my_atoi (lf);
			if (line < 0) {
				/* Function name specified */
				line = -1;                              
				function = lf;
			}
			else
				/* Line number specified */
				function = NULL;
		}
	}

    /* Eexecute the command */
    return fx_info_line (tid, file, line, function);
}


/* ================================================================= 
 * Processes the "info-program <tid>" command.
 */
PUBLIC tkout_t *
cmd_info_program (int argc, const char *argv[])
{
    int tid;

    /* Check arguments */
    if (argc != 2) return f0s_tkout_new (CE_INVALID_NARGS);
    tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Eexecute the command */
    return fx_info_program (tid);
}


/* ================================================================= 
 * Processes the "info-stack <tid>" command.
 */
PUBLIC tkout_t *
cmd_info_stack (int argc, const char *argv[])
{
    int tid;

    /* Check arguments */
    if (argc != 2) return f0s_tkout_new (CE_INVALID_NARGS);
    tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Eexecute the command */
    return fx_info_stack (tid);
}


/* ================================================================= 
 * Processes the "up <tid>" command.
 */
PUBLIC tkout_t *
cmd_up (int argc, const char *argv[])
{
    int tid;

    /* Check arguments */
    if (argc != 2) return f0s_tkout_new (CE_INVALID_NARGS);
    tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Eexecute the command */
    return fx_up (tid);
}


/* ================================================================= 
 * Processes the "down <tid>" command.
 */
PUBLIC tkout_t *
cmd_down (int argc, const char *argv[])
{
    int tid;

    /* Check arguments */
    if (argc != 2) return f0s_tkout_new (CE_INVALID_NARGS);
    tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);
	CHECK_N_ARGS (2);

    /* Eexecute the command */
    return fx_down (tid);
}


/* ================================================================= 
 * Processes the "frame <tid> <frame>" command. 
 * Frame a breakpoint
 */
PUBLIC tkout_t *
cmd_frame (int argc, const char *argv[])
{
    int tid;
    int frame;

    /* Check arguments */
	CHECK_N_ARGS (3);
    frame = my_atoi (argv[2]);
    if (frame < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Eexecute the command */
    return fx_frame (tid, frame);
}


/* ================================================================= 
 * Processes the "info-break <tid>" command.
 */
PUBLIC tkout_t *
cmd_info_break (int argc, const char *argv[])
{
    int tid;

    /* Check arguments */
	CHECK_N_ARGS (2);

    /* Eexecute the command */
    return fx_info_break (tid);
}


/* ================================================================= 
 * Processes the "info-display <tid>" command.
 */
PUBLIC tkout_t *
cmd_info_display (int argc, const char *argv[])
{
    int tid;

    /* Check arguments */
	CHECK_N_ARGS (2);

    /* Eexecute the command */
    return fx_info_display (tid);
}


/* ================================================================= 
 * Processes the "info-threads <tid>" command.
 */
PUBLIC tkout_t *
cmd_info_threads (int argc, const char *argv[])
{
    int tid;

    /* Check arguments */
	CHECK_N_ARGS (2);

    /* Eexecute the command */
    return fx_info_threads (tid);
}


/* ================================================================= 
 * Processes the "thread <tid> <th_id>" command.
 */
PUBLIC tkout_t *
cmd_thread (int argc, const char *argv[])
{
    int tid, th_id;

    /* Check arguments */
	CHECK_N_ARGS (3);
    th_id = my_atoi (argv[2]);
    if (th_id < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

    /* Eexecute the command */
    return fx_thread (tid, th_id);
}


/* ================================================================= 
 * Processes the "sendto <tid> <command>" command. 
 * Send a command directly to the node debugger
 */
PUBLIC tkout_t *
cmd_sendto (int argc, const char *argv[])
{
    char cmd[MAX_BUF];
    int i;
    int tid;
    tkout_t *tk;

    /* Check arguments */
    if (argc < 3) return f0s_tkout_new (CE_INVALID_NARGS);
    tid = my_atoi (argv[0]);
    if (tid < 0) return f0s_tkout_new (CE_INVALID_ARGUMENT);

	/* Join the 3rd and following arguments into just one string */
	strcpy (cmd, "");
	for (i = 2;  i < argc;  i++) {
		strcat (cmd, argv[i]);
		strcat (cmd, " ");
	}

    /* Eexecute the command */
    tk = fx_sendto (tid, cmd);
    return tk;
}
